# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import time
import gtk
import gobject

import emesenelib.common
import desktop

class FileTransfer:
    ''' this class represents one filetransfer'''
    
    def __init__(self, controller, p2p, conversation, session, context, sender):
        '''Constructor
        The sender parameter accepts the special value "Me", which is not
        a valid mail, and is used for sent files'''

        #constants
        #self.state should always be one of those states
        self.WAITING = 0
        self.TRANSFERING = 1
        self.RECEIVED = 2
        self.FAILED = 3
        
        self.controller = controller
        self.p2p = p2p
        self.conversation = conversation
        self.session = int(session)
        self.context = context
        self.sender = sender
        
        self.receivedBytes = 0
        self.state = self.WAITING
        
        self.timeAccepted = None
        
        self.previewImage = None
        if context.preview != '':
            try:
                loader = gtk.gdk.PixbufLoader()
                loader.write(context.preview)
                loader.close()
                
                self.previewImage = loader.get_pixbuf()
                del loader
            except gobject.GError:
                del loader
        
        self.localPath = None

        if sender == 'Me':
            self.conversation.appendOutputText(None, \
                _('Sending %s') % self.context.filename, 'information')
        else:
            args = {'mail': self.p2p.mail, 'file': self.context.filename}
            self.conversation.appendOutputText(None, \
                _('%(mail)s is sending you %(file)s') % args, \
                'information')
        
        self.signals = []
        sap = self.signals.append
        sap(self.p2p.connect('transfer-progress', self.onFtProgress))
        sap(self.p2p.connect('transfer-failed', self.onFtFailed))
        sap(self.p2p.connect('file-transfer-complete', self.onFtReceived))
        if sender == 'Me':
            sap(self.p2p.connect('file-transfer-accepted', self.onFtAccepted))
    
    def disconnect(self):
        for identifier in self.signals:
            self.p2p.disconnect(identifier)
    
    def getPreviewImage(self):
        return self.previewImage
    
    def getBytes(self):
        '''returns a tuple with received and total bytes'''
        return int(self.receivedBytes), int(self.context.file_size)
    
    def accept(self):
        '''accept this transfer'''

        self.conversation.appendOutputText(None, \
            _('You have accepted %s') % self.context.filename, \
            'information')

        self.p2p.emit('file-transfer-accepted', self.session, self.context,
            self.sender)
        self.timeAccepted = time.time()
        
        self.state = self.TRANSFERING
        self.ui.stateChanged()

    def cancel(self):
        '''cancel this transfer'''

        self.conversation.appendOutputText(None,
            _('You have canceled the transfer of %s') % self.context.filename,
            'information')
        
        self.p2p.emit('file-transfer-canceled', self.session, self.context,
            self.sender)

        self.state = self.FAILED
        self.ui.stateChanged()
        self.remove()
    
    def getElapsedTime(self):
        '''return the time elapsed since this transfer was accepted'''
        if not self.timeAccepted:
            return 0
        else:
            return int(time.time()) - int(self.timeAccepted)
    
    def getAverageSpeed(self):
        '''get average file transfer speed (bytes per second)'''
        time = self.getElapsedTime()
        if time == 0: # prevent division by zero
            return 0
        else: 
            return self.receivedBytes / time
    
    def remove(self):
        '''remove this transfer'''
        self.conversation.transfers.remove(self)
        self.disconnect()
      
    def onFtAccepted(self, p2p, session, context, sender):
        if session != self.session: return
        
        self.conversation.appendOutputText(None, \
            _('Starting transfer of %s') % self.context.filename, \
            'information')

        self.timeAccepted = time.time()
        
        self.state = self.TRANSFERING
        self.ui.stateChanged()
        
    def onFtProgress(self, switchboard, session, bytes):
        if session != self.session: return

        self.receivedBytes = int(bytes)
        self.ui.updateProgress()

    def onFtFailed(self, switchboard, session, reason):
        if session != self.session: return
        
        output = _('Transfer of %s failed.') % self.context.filename + ' ' 

        if reason == 'corrupt':
            output += _('Some parts of the file are missing')
        elif reason == 'cancelled':
            output += _('Interrupted by user')
        elif reason == 'error':
            output += _('Transfer error')
        self.conversation.appendOutputText(None, output, 'error')

        self.state = self.FAILED
        self.ui.stateChanged()
        
    def onFtReceived(self, p2p, session, context, src, sender):
        '''called when filetransfer is finished'''
        if session != self.session: return
        
        if sender == 'Me':
            self.conversation.appendOutputText(None, \
                _('%s sent successfully') % context.filename, 'information')
        else:
            self.conversation.appendOutputText(None, \
                _('%s received successfully.') % context.filename, \
                'information')
        self.conversation.doMessageWaiting(sender)
        
        self.receivedBytes = context.file_size
        self.state = self.RECEIVED
        self.ui.stateChanged()
        
        if sender == 'Me' or src.closed:
            return

        # get file dir
        config = self.controller.config
        receivedFilesDir = os.path.expanduser(config.user['receivedFilesDir'])
        
        # if directory is invalid, save to home directory
        if not os.path.exists(receivedFilesDir):
            print receivedFilesDir + ' does not exist. ' \
                'Saving files to home directory.'
            receivedFilesDir = os.path.expanduser('~/')
          
        # separated folders between contacts
        if config.user['receivedFilesSortedByUser']:
            receivedFilesDirSub = os.path.join(receivedFilesDir, p2p.mail)
            
            if os.path.exists(receivedFilesDirSub):
                receivedFilesDir = receivedFilesDirSub
            else:
                os.mkdir(receivedFilesDirSub)
                # ugly, but "if os.mkdir(..)" doesn't work :S
                if os.path.exists(receivedFilesDirSub):
                    receivedFilesDir = receivedFilesDirSub
                    # if we can't create the dir, we put the file directly 
                    # in received files dir
        
        name = os.path.join(receivedFilesDir, self.getFilename())
        
        num = 0
        numstr = ''
        
        while os.path.exists(name + numstr):
            num += 1
            numstr = '.' + str(num)

        self.localPath = name + numstr
        
        # we don't copy because the permissions may be too restrictive
        # and not match the user preferred umask blah blah
        dest = open(self.localPath, 'wb')
        src.seek(0)
        buffer = src.read(32 * 1024)
        while buffer:
            dest.write(buffer)
            buffer = src.read(32 * 1024)
        dest.close()
        src.close()
    
    def open(self):
        '''open received file'''
        if self.localPath != None:
            try:
                desktop.open(self.localPath)
            except OSError:
                pass
    
    def getFilename(self):
        return self.context.filename
    
    def setBytesReceived(self, bytes):
        self.receivedBytes = bytes
        
    def getFraction(self):
        '''return the progress of this transfer as a float from 0 to 1'''
        if self.context.file_size:
            return float(self.receivedBytes) / float(self.context.file_size)
        else:
            return 1



# -*- coding: utf-8 -*-
'''a module to handle contacts'''

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
import gettext

import stock
import status
import Object
import validators

_ = gettext.gettext

class ContactManager(Object.Object):
    '''this class represent an abstract class that provide methods
    to interact with contacts, implementing this class with the
    undeliying protocol library let the upper layer use the services
    of the protocol implementation independently of the api, in this
    way, the protocol library can be modified (or replaced) 
    without effect on the clients'''

    def __init__(self, dialog, protocol, account):
        '''initialize the object, dialog is a implementation
        ot dialog, it's used to interact with the user'''
        Object.Object.__init__(self)

        self.dialog = dialog
        self.protocol = protocol
        self.contacts = {}

        # the object that contains out information
        self.me = Contact(account)
        
        # emitted when some attribute of the contact changes
        # the first argument is the contact object, the second
        # is a string representing the attribute that changed
        # the third is the old value of the attr
        self.signal_add('contact-changed', 3)
        # the second parameter is the old value
        self.signal_add('contact-nick-changed', 2)
        self.signal_add('contact-alias-changed', 2)
        self.signal_add('contact-message-changed', 2)
        self.signal_add('contact-status-changed', 2)
        self.signal_add('contact-picture-changed', 2)
        self.signal_add('contact-media-changed', 2)
        self.signal_add('contact-blocked-changed', 2)

        # contact
        self.signal_add('contact-added', 1)
        # contact
        self.signal_add('contact-removed', 1)

    def register(self, contact):
        '''this method add the contact to the list of contacts'''

        if not self.exists(contact.account):
            self.contacts[contact.account] = contact
        else:
            debug("contact %s already in self.contacts" % (contact.account,))

    # utility methods

    def exists(self, account):
        '''check if the account is on self.contacts, return True if exists'''
        if account in self.contacts:
            return True
        else:
            return False

    # be carefull and don't redefine this methods!

    def _on_contact_added(self, account, identifier=None, nick=None, 
        message='', _status=status.OFFLINE, alias='', blocked=False):
        '''callback called when a new contact is added'''
        if account in self.contacts:
            debug("contact %s already on contacts" % (account,))
        else:
            if not nick:
                nick = account
            contact = Contact(account, identifier, nick, message, _status, 
                    alias, blocked)
            self.register(contact)
            self.signal_emit('contact-added', contact)

    def _on_contact_removed(self, account):
        '''callback called when a contact is removed'''
        if account in self.contacts:
            contact = self.contacts[account]
            del self.contacts[account]
            self.signal_emit('contact-removed', contact)
        else:
            debug("contact %s not in contacts" % (account,))

    def _on_contact_attr_changed(self, account, attr, value):
        '''callback called when an attribute is changed on a contact,
        this attribute can be block or alias (since these are the
        attributes that the user can change, the others are changed
        by the user itself)'''

        if not self.exists(account):
            debug("contact '%s' not in self.contacts" % (account,))
            return

        contact = self.contacts[account]

        if attr == 'block':
            old = contact.blocked
            contact._on_blocked_changed(value)    
            self.signal_emit('contact-blocked-changed', contact, old)
        elif attr == 'alias':
            old = contact.alias
            contact._on_alias_changed(value)    
            self.signal_emit('contact-alias-changed', contact, old)
        else:
            old = contact.attrs.get(attr, None)

        self.signal_emit('contact-changed', contact, attr, old)

    def _on_contact_nick_changed(self, account, nick):
        '''callback called when an user change his nick'''
        if not self.exists(account):
            debug("contact '%s' not in self.contacts" % (account,))
            return

        contact = self.contacts[account]
        old = contact.nick
        contact._on_nick_changed(nick)
        self.signal_emit('contact-changed', contact, 'nick', old)
        self.signal_emit('contact-nick-changed', contact, old)
    
    def _on_contact_message_changed(self, account, message):
        '''callback called when an user change his message'''
        if not self.exists(account):
            debug("contact '%s' not in self.contacts" % (account,))
            return

        contact = self.contacts[account]
        old = contact.message
        contact._on_message_changed(message)
        self.signal_emit('contact-changed', contact, 'message', old)
        self.signal_emit('contact-message-changed', contact, old)
    
    def _on_contact_status_changed(self, account, new_status):
        '''callback called when an user change his status'''
        if not self.exists(account):
            debug("contact '%s' not in self.contacts" % (account,))
            return

        if status.is_valid(new_status):
            contact = self.contacts[account]
            old = contact.status
            contact._on_status_changed(new_status)
            self.signal_emit('contact-changed', contact, 'status', old)
            self.signal_emit('contact-status-changed', contact, old)
        else:
            debug("invalid status '%s'" % (new_status,))
    
    def _on_contact_added_to_group(self, account, group_name):
        '''callback called when an account is added to a group'''
        if self.exists(account):
            if group_name not in self.contacts[account].groups:
                self.contacts[account].groups.append(group_name)
            else:
                debug("group '%s' already in account.groups" % (group_name,))
        else:
            debug("account %s not in self.contacts" % (account,))
            
    def _on_contact_removed_from_group(self, account, group_name):
        '''callback called when an account is removed from a group'''
        if self.exists(account):
            if group_name in self.contacts[account].groups:
                self.contacts[account].groups.remove(group_name)
            else:
                debug("group '%s' not in account.groups" % (group_name,))
        else:
            debug("account %s not in self.contacts" % (account,))

    # actions on our contact
    def get_no_group(self):
        '''return a lost of contacts that dont belong to any group'''
        return [contact for contact in self.contacts.values() \
            if not contact.groups]

    def get_contacts(self, accounts):
        '''return a list of contact objects from a list of accounts'''
        return [self.contacts[account] for account in accounts if account \
            in self.contacts]

    def get_sorted_list_by_status(self, contacts=None):
        '''return a dict with status.* (OFFLINE, ONLINE etc) as key and
        a list of contact objects as value, you can use then 
        status.ORDERED to cycle over the keys.
        The contacts are sorted inside the status by display_name.
        if contacts is None, then use the internal list of contacts
        contacts should be a list of contact objects'''
        sorted_dict = {}
        contacts = contacts or self.contacts.values()

        for stat in status.ORDERED:
            sorted_dict[stat] = [contact for contact in contacts \
                if contact.status == stat]

            sorted_dict[stat].sort(cmp=lambda x, y: cmp(x.display_name, 
                y.display_name))

        return sorted_dict

    def get_sorted_list_by_group(self, groups, sort_by_status=False):
        '''return a dict with group names as keys and a list of sorted
        contacts as value, sort them according to display_name if
        sort_by_status is False, and by status and display_name if
        it's True'''
        groups.sort()
        sorted_dict = {}

        for group in groups:
            contacts = [contact for contact in self.contacts.values() \
                if group in contact.groups]

            if sort_by_status:
                sorted_dict[group] = self.get_sorted_list_by_status(contacts)
            else:
                contacts.sort(cmp=lambda x, y: cmp(x.display_name, 
                    y.display_name))
                sorted_dict[group] = contacts

        return sorted_dict

    def get_online_list(self, contacts=None):
        '''return a list of non offline contacts'''
        contacts = contacts or self.contacts.values()
        
        return [contact for contact in contacts \
                if contact.status != status.OFFLINE]

    def get_online_total_count(self, contacts):
        '''return a tuple with two values, the first is the number of
        non offline contacts on the list, the secont is the total number
        of contacts'''
        total = len(contacts)
        online = len([contact for contact in contacts \
            if contact.status != status.OFFLINE])

        return (online, total)

    def set_nick(self, nick):
        '''set the nick of our account to nick, dont forget to emit
        the contact-nick-changed and contact-changed signals with
        self.me as contact'''
        pass
    
    def get_nick(self, account=None):
        '''return the nick of the account, if account is None, return
        our nick'''
        if account:
            if self.exists(account):
                return self.contacts[account].nick
            else:
                debug("contact '%s' not in self.contacts" % (account,))
                return ''
        else:
            return self.me.nick

    def set_message(self, personal_message):
        '''set the personal message of our account to personal_message, 
        dont forget to emit
        the contact-message-changed and contact-changed signals with
        self.me as contact'''
        pass
    
    def get_message(self, account=None):
        '''return the personal message if account, an empty string if account
        doesn't exist, our message if account=None'''
        if account:
            if self.exists(account):
                return self.contacts[account].message
            else:
                debug("contact '%s' not in self.contacts" % (account,))
                return ''
        else:
            return self.me.message

    def set_media(self, media):
        '''set the current media on our account to media
        dont forget to emit
        the contact-media-changed and contact-changed signals with
        self.me as contact'''
        pass
    
    def get_media(self, account=None):
        '''return the media if account, an empty string if account
        doesn't exist, our media if account=None'''
        if account:
            if self.exists(account):
                return self.contacts[account].media
            else:
                debug("contact '%s' not in self.contacts" % (account,))
                return ''
        else:
            return self.me.media
    
    def set_status(self, new_status):
        '''set the status to status, the status should be one of the
        constants on status.py, consider calling status.is_valid.
        Also you should convert it to the values on the library
        dont forget to emit
        the contact-status-changed and contact-changed signals with
        self.me as contact'''
        pass
    
    def get_status(self, account=None):
        '''return the status of an account if exist, status.OFFLINE if dont
        if account == None, return the status of our user'''
        pass

    def set_picture(self, path):
        '''set the display picture to path'''
        pass

    def get_picture(self, account=None):
        '''return the picture of account, return the picture of the user
        if account is None, the type returned is a string with the content
        of the image'''
        #TODO: implement image cache here
        if account:
            if self.exists(account):
                path = self.contacts[account].picture
            else:
                debug("contact '%s' not in self.contacts" % (account,))
                return None
        else:
            path = self.me.picture

        if validators.readable(path):
            return file(path, 'r').read()
        else:
            return None

    # actions on other contacts
    def set_alias(self, account, alias):
        '''set the contact alias, give an empty alias to reset
        dont forget to emit
        the contact-alias-changed and contact-changed signals'''
        pass
    
    def get_alias(self, account=None):
        '''return the alias of the account, if account is None, return
        our alias'''
        if account:
            if self.exists(account):
                return self.contacts[account].alias
            else:
                debug("contact '%s' not in self.contacts" % (account,))
                return ''
        else:
            return self.me.alias

    def get_display_name(self, account=None):
        '''return the alias or the nick, if account is None, return our
        alias or nick'''
        if account:
            if self.exists(account):
                return self.contacts[account].display_name
            else:
                debug("contact '%s' not in self.contacts" % (account,))
                return ''
        else:
            return self.me.display_name

    def block(self, account):
        '''block an user
        dont forget to emit
        the contact-blocked-changed and contact-changed signals'''
        pass
    
    def get_blocked(self, account):
        '''return True if blocked'''
        if self.exists(account):
            return self.contacts[account].blocked
        else:
            debug("contact '%s' not in self.contacts" % (account,))
            return False
    
    def unblock(self, account):
        '''unblock an user
        dont forget to emit
        the contact-blocked-changed and contact-changed signals'''
        pass

    def remove(self, account):
        '''remove an user'''
        pass
    
    def add(self, account, group=''):
        '''add an user'''
        pass

    def move_to_group(self, account, src_group, dest_group):
        '''move a user from src_group to dest_group'''
        pass

    def add_to_group(self, account, group):
        '''add an user to a group, return True on success'''
        pass
    
    def remove_from_group(self, account, group):
        '''remove an user from a group'''
        pass

    # dialog (more can be added if needed)
    def set_nick_dialog(self):
        '''show a dialog asking to change the nick'''
        self.dialog.set_nick(self.get_nick(), self.set_nick_cb)

    def set_message_dialog(self, old_personal_message):
        '''show a dialog asking to change the personal message'''
        self.dialog.set_message(old_personal_message, 
            self.set_message_cb)

    def set_alias_dialog(self, account):
        '''show the alias dialog'''
        self.dialog.set_contact_alias(account, self.get_alias(account), 
            self.set_alias_cb)

    def set_picture_dialog(self, picture_path, picture_dir):
        '''show a dialog to select a new picture, picture_path
        is the actual picture image path and picture_dir is a
        path to a dir where all the pictures to select are located'''
        self.dialog.set_picture(picture_path, picture_dir,
            self._set_picture_cb)

    def remove_dialog(self, account):
        '''show a confirmation dialog to ask if sure to remove the
        user, it's optional to use, but recomended'''
        self.dialog.yes_no(
            _("Are you sure you want to delete %s?") % (account, ),
            self.remove_cb, account)
    
    def add_dialog(self):
        '''show a dialog to ask for the account, and if the account
        is valid, add the user'''
        self.dialog.add_contact(self.add_cb)

    # callbacks, (the extra parameters are set as '' by default because
    # when the user close the window pressing the x then this parameters
    # are not sent
    def set_nick_cb(self, response, old_nick='', new_nick=''):
        '''callback for the dialog.set_nick method'''
        if response == stock.ACCEPT:
            if old_nick == new_nick:
                debug('old nick and new nick are the same')
                return
            elif new_nick == '':
                debug('empty new nick')
                return

            self.set_nick(new_nick)

    def set_message_cb(self, response, old_pm='', new_pm=''):
        '''callback for the dialog.set_message method'''
        if response == stock.ACCEPT:
            if old_pm == new_pm:
                debug('old and new personal messages are the same')
                return

            self.set_message(new_pm)

    def set_alias_cb(self, response, account='', old_alias='', new_alias=''):
        '''callback for the dialog.set_contact_alias method,
        the parameters and the values are described on that method'''
        if response == stock.ACCEPT:
            if old_alias == new_alias:
                debug('old alias and new alias are the same')
                return

            self.set_alias(account, new_alias)
        elif response == stock.CLEAR:
            self.set_alias(account, '')

    def remove_cb(self, response, account=''):
        '''callback for dialog.yes_no, asking to confirm the 
        user remove'''
        if response == stock.YES:
            self.remove(account)

    def add_cb(self, response, account='', groups=''):
        '''callback to the add_dialog method, add the user and add him 
        to the defined groups'''
        if response == stock.ADD:
            self.add(account)

            if groups:
                for group in groups:
                    self.add_to_group(account, group)

    def _set_picture_cb(self, response, path):
        '''callback to the set_picture_dialog, change the picture of
        self.me'''
        if response == stock.ACCEPT:
            self.set_picture(path)
        elif response == stock.CLEAR:
            self.set_picture('')

class Contact(object):
    '''a class that represent a contact'''

    def __init__(self, account, identifier=None, nick='', message=None,
        _status=status.OFFLINE, alias='', blocked=False):
        '''class contructor'''
        self.account = account
        self.identifier = identifier or '0'
        self.nick = nick or self.account
        # message is the personal message or status message, as you may
        # call it
        self.message = message or ''
        self.media = ''
        self.status = _status
        self.alias = alias
        self.blocked = blocked
        self.picture = None
        self.groups = []

        # extra atributes (use contact.attrs.get("attr", "default"))
        self.attrs = {}

    def _on_nick_changed(self, nick):
        '''callback called on nick change'''
        self.nick = nick

    def _on_message_changed(self, message):
        '''callback called on message change'''
        self.message = message

    def _on_status_changed(self, _status):
        '''callback called on status change'''
        self.status = _status

    def _on_alias_changed(self, alias):
        '''callback called on alias change'''
        self.alias = alias

    def _on_blocked_changed(self, blocked):
        '''callback called on blocked change'''
        self.blocked = blocked

    def _get_display_name(self):
        '''return the alias if exist, if not the nick if not empty, if not
        the mail'''

        return self.alias or self.nick or self.account

    display_name = property(fget=_get_display_name)

    def _get_status_string(self):
        '''return a string representation of the status'''
        return status.STATUS.get(self.status, '?')

    status_string = property(fget=_get_status_string)

    def __repr__(self):
        '''return a string representation of the object'''
        return "<contact account='%s'>" % (self.account,)

def debug(msg):
    '''debug method, the module send the debug here, it can be changed
    to use another debugging method'''
    return
    print('ContactManager.py: ', msg)



# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

'''module that implements abstract.ConfigDialog using Gtk'''

import gtk

from abstract.ConfigDialog import *
import dialog

BUILDERS = {}

def build_generic(element):
    '''build and return the element'''
    if type(element) in BUILDERS:
        return BUILDERS[type(element)](element)
    else:
        raise ValueError("unknown type of element")

def _text_cb(widget, element):
    '''method called when the value of the widget is changed, set the value
    of the text to the element'''
    element.value = widget.get_text()

def _checkbox_cb(widget, element):
    '''method called when the element is toggled, change the value of the 
    element'''
    element.value = widget.get_active()

def _radio_cb(widget, element):
    '''method called when a radio element is toggled, change the selected
    value to the identifier of that radio'''
    if not widget.get_active():
        return

    index = list(element.labels).index(widget.get_label())
    element.value = element.identifiers[index]

def build_text(element, spacing=2, border=2, visibility=True):
    '''build a Text object'''
    hbox = gtk.HBox(spacing=spacing)
    hbox.set_border_width(border)
    label = gtk.Label(element.label)
    entry = gtk.Entry()
    entry.set_text(element.value)
    entry.set_visibility(visibility)
    entry.connect("changed", _text_cb, element)

    hbox.pack_start(label, True, True)
    hbox.pack_start(entry, True, True)

    return hbox

def build_password(element, spacing=2, border=2):
    '''build a Text object'''
    return build_text(element, spacing, border, False) 

def build_checkbox(element):
    '''build a checkbox'''
    checkbox = gtk.CheckButton(element.label)
    checkbox.set_active(element.value)
    checkbox.connect("toggled", _checkbox_cb, element)

    return checkbox

def build_radiogroup(element, spacing=2, border=2):
    '''build a radio group'''
    first_radio = gtk.RadioButton(None, element.labels[0])
    first_radio.connect("toggled", _radio_cb, element)
    vbox = gtk.VBox(spacing=spacing)
    vbox.set_border_width(border)
    vbox.pack_start(first_radio, True, True)

    if element.selected_index == 0:
        first_radio.set_selected(True)

    for (index, label) in enumerate(element.labels[1:]):
        radio = gtk.RadioButton(first_radio, label)
        radio.connect("toggled", _radio_cb, element)

        if index + 1 == element.selected_index:
            radio.set_active(True)

        vbox.pack_start(radio, True, True)

    return vbox

def build_group(element, spacing=2, border=2):
    '''build a group'''
    vbox = gtk.VBox(spacing=spacing)
    vbox.set_border_width(border)

    for elem in element.items:
        vbox.pack_start(build_generic(elem))

    if element.label:
        frame = gtk.Frame(element.label)
        frame.add(vbox)
        return frame
    else:
        return vbox

def build_tab(element, spacing=2, border=2):
    '''build a tab'''
    vbox = gtk.VBox(spacing=spacing)
    vbox.set_border_width(border)

    for elem in element.items:
        vbox.pack_start(build_generic(elem))

    return vbox

def build_tabgroup(element, spacing=2, border=2):
    '''build a tabgroup'''
    notebook = gtk.Notebook()

    for tab in element.items:
        notebook.append_page(build_generic(tab), gtk.Label(tab.label))

    return notebook

def _ok_clicked_cb(button, window, element):
    '''called when the ok button is clicked'''
    if _on_close(element):
        window.hide()

def _close_cb(window, event, element):
    '''called when the dialog is closed, element is the root container
    will call to on_last_change to inform that is the last change'''
    if _on_close(element):
        window.hide()

    return True # stop the window from being destroyed (you can reuse the 
    # window if you mantain a reference to it)

def _on_close(element):
    '''calidate the fields and display a dialog if something is not 
    valid and return False, if all validated, return True'''
    (validated, message, element) = element.validate()

    if not validated:
        tmp = {
            'identifier': element.identifier,
            'value': element.value,
            'message': message,
        }
        dialog.error(_("Field '%(identifier)s ('%(value)s') not valid\n" \
            "'%(message)s'" % tmp))
        return False
    else:
        element.on_last_change()
        return True

def build(element, spacing=2, border=2):
    '''build a window and return it'''

    window = gtk.Window()
    vbox = gtk.VBox(spacing=spacing)
    vbox.set_border_width(border)

    button_box = gtk.HButtonBox()
    button_box.set_layout(gtk.BUTTONBOX_END)
    ok_button = gtk.Button(stock=gtk.STOCK_OK)
    ok_button.connect("clicked", _ok_clicked_cb, window, element)
    button_box.pack_start(ok_button)

    vbox.pack_start(build_generic(element), True, True)
    vbox.pack_start(button_box, False)

    window.add(vbox)
    vbox.show_all()
    window.connect("delete-event", _close_cb, element)

    return window

BUILDERS[Text] = build_text
BUILDERS[Password] = build_password
BUILDERS[CheckBox] = build_checkbox
BUILDERS[RadioGroup] = build_radiogroup
BUILDERS[Group] = build_group
BUILDERS[Tab] = build_tab
BUILDERS[TabGroup] = build_tabgroup



# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import md5
import common
from time import time

class Hotmail:
    def __init__( self, msn, config ):
        self.user = msn.user
        self.password = common.escape(msn.password)
        self.MSPAuth = msn.MSPAuth
        self.dg = msn.demographics
        self.config = config
        
    def getLoginPage( self, MessageURL=None , PostURL=None, id='2' ):
        if PostURL == None:
            PostURL = 'https://loginnet.passport.com/ppsecure/md5auth.srf?lc='+self.dg['lang_preference']
        if MessageURL == None:
            MessageURL = "/cgi-bin/HoTMaiL"
           
        sl = str( int ( time() ) - int( self.dg['LoginTime'] ) )
        auth = self.MSPAuth
        sid = self.dg['sid']
        cred =  md5.new( auth + sl + self.password ).hexdigest()

        templateData = {
        'id':id,
        'site':PostURL,
        'login': self.user.split('@')[0],
        'email':self.user,
        'sid':sid,
        'kv':'',
        'sl':sl,
        'url':MessageURL,
        'auth':auth,
        'creds':cred
        }
        
        return self.parseTemplate( templateData )
        
    def getProfilePage( self, user ):
        pass
        
    def parseTemplate( self, data ):
        f = open( 'hotmlog.htm' )
        hotLogHtm = f.read()
        f.close()
        for key in data:
            hotLogHtm = hotLogHtm.replace( '$'+key, data[ key ] )

        self.file = os.path.join(
            self.config.getUserConfigPath(), 'cache', 'login.htm')
        
        tmpHtml = open( self.file, 'w' )
        tmpHtml.write( hotLogHtm )
        tmpHtml.close()
        
        return 'file:///' + self.file
        



# -*- coding: utf-8 -*-

#   This file is a plugin for emesene.
#
#    Dbus Emesene Plugin is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    Dbus Emesene Plugin is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import Plugin
from gobject import timeout_add, source_remove

try:
    import dbus
    dbusError = ''
    if getattr( dbus, 'version', ( 0, 0, 0 ) ) >= ( 0, 41, 0 ):
        import dbus.glib
except Exception, e:
    dbusError = e
    
class MainClass( Plugin.Plugin ):
    
    def __init__( self, controller, msn ):
        Plugin.Plugin.__init__( self, controller, msn )
        self.description = _( 'Changes status to selected one if session is idle (you must use gnome and have gnome-screensaver installed)' ) 
        self.authors = { 'mg' : 'themgzzy at gmail dot com' }
        self.displayName = _( 'Idle Status Change' )
        self.name = 'IdleStatusChange'
        
        self.config = controller.config
        self.config.readPluginConfig( self.name )
        
        self.idlestatus = self.config.getPluginValue( self.name, 'idlestatus', '' )
        
        self.controller = controller
        self.msn = msn

        self.oldstatus = self.msn.status
        self.isIdle = False
        
    def start( self ):
        '''start the plugin'''
        try:
            self.startDbus()
        except:
            return
        self.enabled = True
        self.tag = timeout_add( 5000, self.idle_state )
    
    def stop( self ):
        source_remove(self.tag)
        self.enabled = False
        
    def check( self ):
        '''
        check if everything is OK to start the plugin
        return a tuple whith a boolean and a message
        if OK -> ( True , 'some message' )
        else -> ( False , 'error message' )
        '''
        if dbusError != '':
            return ( False, 'Can\'t Load dbus ' + str(dbusError) )
            
        return ( True, 'Ok' )
    
    def configure( self ):
        statuslist = [ 'idle', 'online', 'away', 'busy', 'brb', 'phone', 
             'lunch', 'invisible' ]
        l = [ Plugin.Option( 'idlestatus', list, _( 'Idle Status:' ), _( 'Set the idle status' ), self.config.getPluginValue( self.name, 'idlestatus', '' ), statuslist ) ]
        response = Plugin.ConfigWindow( _( 'Idle Status Change configuration' ), l ).run()
        
        if response != None:
            self.idlestatus = response['idlestatus'].value
            self.config.setPluginValue( self.name, 'idlestatus', self.idlestatus )
        return True
        
    def startDbus( self ):
        '''Start dbus session'''
        self.session_bus = dbus.SessionBus()
        self.screensaver = self.session_bus.get_object( 'org.gnome.ScreenSaver', '/org/gnome/ScreenSaver' )
        
    def idle_state( self ):
        idlestate = self.screensaver.GetSessionIdle()
        if idlestate and not self.isIdle and self.msn.status != "HDN":
            self.isIdle = True
            self.oldstatus = self.msn.status
            self.msn.changeStatus( self.idlestatus )

        elif not idlestate and self.isIdle:
            self.isIdle = False
            self.msn.changeStatus( self.oldstatus )

        
        return True



# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import re
import gtk
import gtk.gdk
import pango
import cairo
import gobject

class cairoColor(object):
    '''An object that describes a color using
    red, green, blue attributes'''

    def __init__(self, gdkcolor):
        self.red = float(gdkcolor.red) / 65535
        self.green = float(gdkcolor.green) / 65535
        self.blue = float(gdkcolor.blue) / 65535

class FancyLabel( gtk.Widget ):
    '''Cairo-based label that renders (...) '''
     
    __gproperties__ = {
        'text': (str, '', '', '', gobject.PARAM_READWRITE),
        'markup': (bool, '', '', True, gobject.PARAM_READWRITE),
        'plusmarkup': (bool, '', '', True, gobject.PARAM_READWRITE),
        'smilies': (bool, '', '', True, gobject.PARAM_READWRITE),
        'max-lines': (int, '', '', -1, 10000, -1, gobject.PARAM_READWRITE),
        'max-line-height': (int, '', '', -1, 100, -1, gobject.PARAM_READWRITE),
        'wrap': (bool, '', '', True, gobject.PARAM_READWRITE),
        'wrap-by-char': (bool, '', '', False, gobject.PARAM_READWRITE),
        'ellipsize-end': (bool, '', '', True, gobject.PARAM_READWRITE),
    }

    __gsignals__ = {
        'size_request' : 'override',
        'expose-event' : 'override',
        'size_allocate' : 'override' ,
    }

    #colors taken from plus plugin, codes 00 - 67
    _colorCodes = (
    'ffffff','000000','00007D','009100','FF0000','7D0000','9A009A','FC7D00',
    'FFFF00','00FC00','009191','00FFFF','1E1EFC','FF00FF','7D7D7D','D2D2D2',
    'E7E6E4','cfcdd0','ffdea4','ffaeb9','ffa8ff','c1b4fc','bafbe5','ccffa3',
    'fafda2','b6b4b7','a2a0a1','f9c152','ff6d66','ff62ff','6c6cff','68ffc3',
    '000000','f9ff57','858482','6e6d7b','ffa01e','F92411','FF1EFF','1E29FF',
    '7dffa5','60f913','fff813','5e6464','4b494c','d98812','eb0505','de00de',
    '0000d3','03cc88','59d80d','d4c804','333335','18171c','944e00','9b0008',
    '980299','01038c','01885f','389600','9a9e15','473400','4d0000','5f0162',
    '000047','06502f','1c5300','544d05')

    #giant hackish regular expression :D , a match will be one of these groups
    _bbmarkupRE = re.compile( \
    r'(?P<bOpen>\[b\])|(?P<bClose>\[/b\])|(?P<iOpen>\[i\])|(?P<iClose>\[/i\])|'
    r'(?P<uOpen>\[u\])|(?P<uClose>\[/u\])|(?P<sOpen>\[s\])|(?P<sClose>\[/s\])|'
    r'(\[c=(?P<cOpen>([\w]{3,20})|(#([0-9a-f]{6,6}))|([\d]{1,2}))\])|'   
    r'(\[/c(?P<cClose>(=[\w]{3,20})|(=#([0-9a-f]{6,6}))|(=[\d]{1,2})|())\])|'
    r'(\[a=(?P<aOpen>([\w]{3,20})|(#([0-9a-f]{6,6}))|([\d]{1,2}))\])|'
    r'(\[/a(?P<aClose>(=[\w]{3,20})|(=#([0-9a-f]{6,6}))|(=[\d]{1,2})|())\])|',
    re.IGNORECASE)
    
    _smiliesRE = None
    _smilieDescent = 0.25   # percentage of smilie to use as descent

    _text = ''
    _markup = True          # render pango markup
    _plusmarkup = True      # render emesenePlus markup
    _smilies = True         # render smilies 
    _max_lines = -1         # set to -1 to render all lines
    _max_line_height = -1   # set to -1 for natural height,
                            ## otherwise scale to pixel height
    _wrap = True            # wrap text when true 
                            ## (see wrap_by_char for wrap type)
    _wrap_by_char = False   # False: wrap word, True: wrap characters 
                            ## (ignored if _wrap is False) 
    _ellipsize_end = True   # self explanitory

    # cairo related resources
    _cachedSmilieDict = {}
    _smilieStack = []    
    _foregroundStack = []
    _backgroundStack = []
    _cairoForegroundStack = []
    _cairoBackgroundStack = []
        
    _layout = None
    _ellipsize_width = 0
    _cairo_scale = 1
    _fixed_layout_height = 0    # set on allocate, used for expose

    def __init__(self, controller, text=''):
        gobject.GObject.__init__(self)
        gtk.Widget.__init__(self)
        self.set_flags(self.flags() | gtk.NO_WINDOW )
        self._theme = controller.theme
        self._config = controller.config
        self._smiliesRE = self.__createSmiliesRE(self._theme.getSmileysList())

        self._text = text
        self.__setup_layout()
        
    def do_get_property(self, prop):
        if prop.name == 'text':
            return self._text
        elif prop.name == 'markup':
            return self._markup
        elif prop.name == 'plusmarkup':
            return self._plusmarkup
        elif prop.name == 'smilies':
            return self._smilies
        elif prop.name == 'max-lines':
            return self._max_lines
        elif prop.name == 'max-line-height':
            return self._max_line_height
        elif prop.name == 'wrap':
            return self._wrap
        elif prop.name == 'wrap-by-char':
            return self._wrap_by_char
        elif prop.name == 'ellipsize-end':
            return self._ellipsize_end
        else:
            raise AttributeError, 'unknown property %s' % prop.name
  
    def do_set_property(self, prop, value):
        if prop.name == 'text':
            self._text = value
            self.__setup_layout()
            self.queue_resize()
        elif prop.name == 'markup':
            self._markup = value
            self.__setup_layout()
            self.queue_resize()
        elif prop.name == 'plusmarkup':
            self._plusmarkup  = value
            self.__setup_layout()
            self.queue_resize()  
        elif prop.name == 'smilies':
            self._smilies = value
            self.__setup_layout()
            self.queue_resize()
        elif prop.name == 'max-lines':
            self._max_lines = value
            self.queue_resize()
        elif prop.name == 'max-line-height':
            self._max_line_height = value
            self.__setup_cairo_vars()
            self.queue_resize()
        elif prop.name == 'wrap':
            self._wrap = value
            self.queue_resize()   
        elif prop.name == 'wrap-by-char':
            self._wrap_by_char = value
            self.queue_resize()
        elif prop.name == 'ellipsize-end':
            self._ellipsize_end = value
            self.queue_resize()
        else:
            raise AttributeError, 'unknown property %s' % prop.name

    #Public methods
    def set_text(self, text):
        self.set_property('text', text)
    def set_markup(self, text):
        self.set_property('text', text)
    def get_text(self):
        return self.get_property('text')

    def set_max_line_height(self, height):
        self.set_property('max-line-height', height)
    def get_max_line_height(self):
        return self.get_property('max-line-height')

    def set_use_markup(self, setting):
        self.set_property('markup', setting)
    def get_use_markup(self):
        return self.get_property('markup')

    def set_use_plusmarkup(self, setting):
        self.set_property('plusmarkup', setting)
    def get_use_plusmarkup(self):
        return self.get_property('plusmarkup')

    def set_use_smilies(self, setting):
        self.set_property('smilies', setting)
    def get_use_smilies(self):
        return self.get_property('smilies')

    def set_ellipsize_end(self, setting):
        self.set_property('ellipsize-end', setting)
    def get_ellipsize_end(self):
        return self.get_property('ellipsize-end')

    def set_max_lines(self, maxlines):
        self.set_property('max-lines', maxlines)
    def get_max_lines(self):
        return self.get_property('max-lines')

    def set_wrap(self, setting):
        self.set_property('wrap', setting)
    def get_wrap(self):
        return self.get_property('wrap')
    
    def set_wrap_by_char(self, setting):
        self.set_property('wrap-by-char', setting) 

    def get_wrap_by_char(self):
        return self.get_property('wrap-by-char')
        
    def __compare_attr(self, attribute):
        if attribute.type == pango.ATTR_FOREGROUND: 
            return True
        else:
            return False

    def __setup_layout(self):
        self._layout = self.create_pango_layout('')
        if self._markup:
            self._layout.set_markup(self._text)
        else:
            self._layout.set_text(self._text)
        attributes = self._layout.get_attributes()

        if attributes == None:
            attributes = pango.AttrList()
        
        plainText = self._layout.get_text()
        if self._plusmarkup:
            self.__parse_plusmarkup(attributes, plainText)
        if self._smilies:
            self.__parse_smilies(attributes, plainText)
        self._layout.set_attributes(attributes)
        self.__setup_cairo_vars()


    def __setup_cairo_vars(self):
        # set up _cairo_scale
        line_height = 0
        for i in range(self._layout.get_line_count()):
            current = self._layout.get_line(i).get_pixel_extents()[1][3]
            line_height = max(line_height, current)
        if 0 < self._max_line_height < line_height:
            self._cairo_scale = float(self._max_line_height) / line_height
        else:
            self._cairo_scale = 1
                    
        # set ellipsize width
        ellipsize_layout = self.create_pango_layout('')
        ellipsize_layout.set_markup('<b>...</b>')
        self._ellipsize_width = ellipsize_layout.get_pixel_size()[0]


    def __parse_plusmarkup(self, attributes, text):
        self._foregroundStack = []
        self._backgroundStack = []
        bStack, iStack, uStack, sStack, cStack, aStack = [], [], [], [], [], []
        iterator = self._bbmarkupRE.finditer(text)
        for match in iterator:
            tags = match.groupdict()
            if not tags['bOpen'] == None:
                bStack.append(match.span()) 
            elif not tags['bClose'] == None and len(bStack)> 0:
                openTag, closeTag = bStack.pop(), match.span()
                attributes.insert(pango.AttrWeight(pango.WEIGHT_BOLD, \
                    openTag[1], closeTag[0]))
                self.__hideTag(attributes, openTag)
                self.__hideTag(attributes, closeTag)
            elif not tags['iOpen'] == None:
                iStack.append(match.span())
            elif not tags['iClose'] == None and len(iStack)> 0:
                openTag, closeTag = iStack.pop(), match.span()
                attributes.insert(pango.AttrStyle(pango.STYLE_ITALIC, \
                    openTag[1], closeTag[0]))
                self.__hideTag(attributes, openTag)
                self.__hideTag(attributes, closeTag)
            elif not tags['uOpen'] == None:
                uStack.append(match.span()) 
            elif not tags['uClose'] == None and len(uStack)> 0:
                openTag, closeTag = uStack.pop(), match.span()
                attributes.insert(pango.AttrUnderline(pango.UNDERLINE_SINGLE, \
                    openTag[1], closeTag[0]))
                self.__hideTag(attributes, openTag)
                self.__hideTag(attributes, closeTag)
            elif not tags['sOpen'] == None: sStack.append(match.span()) 
            elif not tags['sClose'] == None and len(sStack)> 0:
                openTag, closeTag = sStack.pop(), match.span()
                attributes.insert(pango.AttrStrikethrough(True, \
                    openTag[1], closeTag[0]))
                self.__hideTag(attributes, openTag)
                self.__hideTag(attributes, closeTag)
            elif not tags['cOpen'] == None:
                cStack.append((match.span(), \
                    self.__colorFromParser(tags['cOpen'])))
            elif not tags['cClose'] == None and len(cStack)> 0:
                openTag, colora = cStack.pop()
                closeTag = match.span()
                if len(tags['cClose']) > 0:
                    colorb = self.__colorFromParser(tags['cClose'])
                else:
                    colorb = colora
                self._foregroundStack.append(( openTag[1], closeTag[0], \
                    cairoColor(colora), cairoColor(colorb) ))
                #attributes.insert(pango.AttrForeground(colora.red, \
                #   colora.green, colora.blue, openTag[1], closeTag[0]))
                self.__hideTag(attributes, openTag)
                self.__hideTag(attributes, closeTag)
            elif not tags['aOpen'] == None:
                aStack.append((match.span(), \
                    self.__colorFromParser(tags['aOpen'])))
            elif not tags['aClose'] == None and len(aStack)> 0:
                openTag, colora = aStack.pop()
                closeTag = match.span()
                if len(tags['aClose']) > 0:
                    colorb = self.__colorFromParser(tags['aClose'])
                else:
                    colorb = colora
                self._backgroundStack.append(( openTag[1], closeTag[0], \
                    cairoColor(colora), cairoColor(colorb) ))
                #attributes.insert(pango.AttrBackground(colora.red, \
                #   colora.green, colora.blue, openTag[1], closeTag[0]))
                self.__hideTag(attributes, openTag)
                self.__hideTag(attributes, closeTag)
    
    def __hideTag(self, attrlist, span):
        attrlist.insert(pango.AttrShape((0, 0, 0, 0), (0, 0, 0, 0), \
            span[0], span[1]))
    
    def __colorFromParser(self, code):  
        codeFix = code.replace('#', '').replace('=', '')
        if len(codeFix)> 2:
            return gtk.gdk.color_parse(codeFix)
        else:
            if int(codeFix) < 68:
                code = self._colorCodes[int(codeFix)]
            else:
                code = self._colorCodes[1]
            return gtk.gdk.color_parse('#' + code)

    def __parse_smilies(self, attributes, plainText):
        
        self._cachedSmilieDict = {}
        self._smilieStack = []    
        
        if self._smiliesRE:
            iterator = self._smiliesRE.finditer(plainText)
        else:
            return

        for match in iterator:
            code = match.group()
            start, end = match.span()            
            if not self._cachedSmilieDict.has_key(code):
                self._cachedSmilieDict[code] = \
                    self._theme.getSmiley(code, False)
            pixbuf = self._cachedSmilieDict[code]
            height, width = pixbuf.get_height() , pixbuf.get_width()
            descent = int(height * self._smilieDescent)
            rect = (0, \
                (descent - height) * pango.SCALE, \
                width * pango.SCALE, \
                height * pango.SCALE)
            attributes.insert(pango.AttrShape( (0, 0, 0, 0), \
                rect, start, start + 1 ))
            if end > (start + 1):
                attributes.insert( \
                    pango.AttrShape((0, 0, 0, 0), (0, 0, 0, 0), start + 1, end))
            self._smilieStack.append((start, code)) # index, key

    def __createSmiliesRE(self, smilieList):

        def compare(x, y):
            if len(x) == len(y):
                return 0
            elif len(x) < len(y):
                return 1
            else:
                return -1
            
        #special characters in re  . ^ $ + ? { } [ ] \ | ( ) *
        special_characters = [r'\.', r'\^', r'\$', r'\+', r'\?', \
            r'\{', r'\}', r'\[', r'\]', r'\\', r'\|', r'\(', r'\)', r'\*']
        items = smilieList[:]
        items.sort(compare)
        items_expression = []
        for item in items:
            for character in special_characters:
                item = re.sub(character, character, item)
            items_expression.append(item)
        return re.compile( '|'.join(items_expression) )
        
    def do_size_request(self, requisition):
        if not self._layout:
            self.__setup_layout()
        self._layout.set_width(-1)
        width, height = self._layout.get_pixel_size()
        width, height = width * self._cairo_scale, height * self._cairo_scale
        
        if not self._wrap and not self._ellipsize_end:
            requisition.width = width

        elif self._wrap and not self._ellipsize_end:
            self._layout.set_width(0)
            if self._wrap_by_char: self._layout.set_wrap(pango.WRAP_CHAR)
            else: self._layout.set_wrap(pango.WRAP_WORD)
            requisition.width = self._layout.get_pixel_size()[0] * \
                self._cairo_scale

        elif not self._wrap and self._ellipsize_end:
            if width < (self._ellipsize_width * self._cairo_scale):
                requisition.width = width
            else:
                requisition.width = self._ellipsize_width * self._cairo_scale

        requisition.height = height
            
    def do_size_allocate(self, allocation):
        scale = self._cairo_scale
        if self._wrap:
            self._layout.set_width( (allocation.width/scale) * pango.SCALE)
            if self._wrap_by_char:
                self._layout.set_wrap(pango.WRAP_CHAR)
            else:
                self._layout.set_wrap(pango.WRAP_WORD)
            request_width = -1
        else:
            self._layout.set_width(-1)
            if self._ellipsize_end:
                request_width = -1
            else:
                request_width = self._layout.get_pixel_extents()[1][2] * scale
        
        height = 0
        for i in range(self.__get_fixed_line_count()):
            height += self._layout.get_line(i).get_pixel_extents()[1][3]
        
        self._fixed_layout_height = height * scale
        self.set_size_request(request_width, self._fixed_layout_height)
        
        self._cairoForegroundStack = self.__create_cairo_gradients( \
            self._foregroundStack)
        self._cairoBackgroundStack = self.__create_cairo_gradients( \
            self._backgroundStack)
            
        gtk.Widget.do_size_allocate(self, allocation)
        
    def __get_gradient_rectangle(self, a, b):
        temp_a = self._layout.index_to_pos(a)
        temp_b = self._layout.index_to_pos(b)
        x = pango.PIXELS(temp_a[0])
        y = pango.PIXELS(temp_a[1])
        width =  pango.PIXELS(temp_b[0] - temp_a[0])
        height = pango.PIXELS(temp_a[3])
        return gtk.gdk.Rectangle(x, y, width, height)
        
    def __create_cairo_gradients(self, stack):
        cairo_stack = []
        for item in stack:
            rectangles = []
            start_i, end_i, color_a, color_b = item
            i = 0
            start_line_found = False
            end_line_found = False
            gradient_width = 0
            for i in range(self._layout.get_line_count()):
                line = self._layout.get_line(i)
                lstart = line.start_index
                if lstart <= start_i < lstart + line.length:
                    start_line_found = True
                    if lstart < start_i <= lstart + line.length:
                        end_line_found = True
                        rect = self.__get_gradient_rectangle(start_i, end_i )
                        gradient_width += rect.width
                        rectangles.append(rect)
                        break
            if not start_line_found:
                break
            elif not end_line_found:
                rect = self.__get_gradient_rectangle(start_i, \
                    line.start_index + line.length )
                gradient_width += rect.width
                rectangles.append(rect)
                for j in range( i + 1 , self._layout.get_line_count() - 1 ):
                    line = self._layout.get_line(j)
                    start = line.start_index
                    if start < end_i <= start + line.length:
                        end_line_found = True
                        rect = self.__get_gradient_rectangle( \
                            line.start_index, end_i )
                        gradient_width += rect.width
                        rectangles.append(rect)
                        break
                    else:
                        rect = self.__get_gradient_rectangle( \
                            line.start_index, line.start_index + \
                            line.length - 1)
                        gradient_width += rect.width
                        rectangles.append(rect)
            if not end_line_found:
                break
            shift = 0
            for rect in rectangles:
                gradient = cairo.LinearGradient(rect.x - shift, rect.y, \
                    rect.x - shift + gradient_width, rect.y)   
                gradient.add_color_stop_rgb(0, \
                    color_a.red, color_a.green, color_a.blue)
                gradient.add_color_stop_rgb(1, \
                    color_b.red, color_b.green, color_b.blue)
                shift += rect.width
                cairo_stack.append((rect, gradient))
        return cairo_stack


    def __line_number_from_start_index(self, index):
        for i in range(self._layout.get_line_count()):
            line = self._layout.get_line(i)
            if line.start_index <= index < line.start_index + line.length:
                return i
        return -1

    def __line_number_from_end_index(self, index):
        for i in range(self._layout.get_line_count()):
            line = self._layout.get_line(i)
            if line.start_index < index <= line.start_index + line.length:
                return i
        return -1 

    def __get_fixed_line_count(self):
        if self._max_lines >= 0:
            return min(self._layout.get_line_count(), self._max_lines)
        else: return self._layout.get_line_count()
        
    def do_expose_event(self, evnt):
        cell_area = self.get_allocation()
        ctx =  self.window.cairo_create()
        ctx.translate(cell_area.x, cell_area.y)

        ctx.scale(self._cairo_scale, self._cairo_scale)
        cairo_width = cell_area.width / self._cairo_scale
        ctx.set_antialias(cairo.ANTIALIAS_SUBPIXEL)
        # set up clip
        if self._ellipsize_end:
            path_coordinates = []
            x = 0
            y = 0
            for i in range(self.__get_fixed_line_count()):  
                line = self._layout.get_line(i) 
                clip_test = line.x_to_index(cairo_width * pango.SCALE)
                char_position = self._layout.index_to_pos(clip_test[1])
                if clip_test[0]: #if cliped
                    clip_test = line.x_to_index((cairo_width * pango.SCALE) - \
                        (self._ellipsize_width * pango.SCALE))
                    char_position = self._layout.index_to_pos(clip_test[1])
                    if clip_test[0]:
                        line_rect = gtk.gdk.Rectangle(0, y, \
                            char_position[0] / pango.SCALE, \
                            char_position[3] / pango.SCALE)
                    else:
                        line_rect = gtk.gdk.Rectangle(0, y, \
                            self._ellipsize_width, \
                                char_position[3] / pango.SCALE)   
                    #paint ellipsize
                    ctx.save()
                    ellipsize_layout = self.create_pango_layout('...')
                    z1 = line.get_pixel_extents()[1][1]
                    z2 = ellipsize_layout.get_line(0).get_pixel_extents()[1][1]
                    ctx.move_to(line_rect.width, (line_rect.y - z1  + z2))
                    ctx.show_layout(ellipsize_layout)
                    ctx.restore()
                else:
                    line_rect = gtk.gdk.Rectangle(0, y, \
                        line.get_pixel_extents()[1][2], \
                        char_position[3] / pango.SCALE)
                x = line_rect.width
                path_coordinates.append((x, y))
                y += line_rect.height
                path_coordinates.append((x, y))
            x = 0
            path_coordinates.append((x, y))
            ctx.move_to(0, 0)
            for coordinate in path_coordinates:
                ctx.line_to(coordinate[0], coordinate[1])
            ctx.close_path()
            ctx.clip()
        else:
            ctx.rectangle(0, 0, cairo_width, self._fixed_layout_height)
            ctx.clip()
        
        #paint background
        for background in self._cairoBackgroundStack:
            rect, gradient = background
            ctx.save() 
            ctx.rectangle(rect.x, rect.y  , rect.width, rect.height)
            ctx.clip_preserve()
            ctx.set_source(gradient)
            ctx.fill()
            ctx.restore()
        #paint middleground
        ctx.save()
        ctx.move_to(0, 0)
        ctx.layout_path(self._layout)
        ctx.clip()
        ctx.show_layout(self._layout)
        ctx.restore()
        
        #paint foreground
        for foreground in self._cairoForegroundStack:
            rect, gradient = foreground
            ctx.save() 
            ctx.rectangle(rect.x, rect.y  , rect.width, rect.height)
            ctx.clip()
            ctx.move_to(0, 0)
            ctx.layout_path(self._layout)
            ctx.set_source(gradient)
            ctx.fill()
            ctx.restore()
        
        #paint smilies
        for smilie in self._smilieStack:
            # remember later to change this so it paints to the
            # base of the line
            index, key = smilie
            x, y, width, height = self._layout.index_to_pos(index)
            if self._cachedSmilieDict.has_key(key):
                pixbuf = self._cachedSmilieDict[key]
                ctx.save()
                #ctx.translate()
                ctx.set_source_pixbuf(pixbuf, x/pango.SCALE, y/pango.SCALE)
                ctx.paint()
                ctx.restore()

gobject.type_register( FancyLabel )



# -*- coding: utf-8 -*-
'''a module to handle abstract status for different IM clients'''

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#import gettext
#_ = gettext.gettext

OFFLINE = 0
ONLINE = 1
BUSY = 2
AWAY = 3
IDLE = 4
INVISIBLE = 5
LUNCH = 6
TELEPHONE = 7
BE_RIGHT_BACK = 8

ORDERED = [ONLINE, BUSY, AWAY, BE_RIGHT_BACK, IDLE, LUNCH, TELEPHONE,
    INVISIBLE, OFFLINE]

STATUS = {OFFLINE : _('Offline'),
    ONLINE : _('Online'),
    BUSY : _('Busy'),
    AWAY : _('Away'),
    IDLE : _('Idle'),
    INVISIBLE : _('Invisible'),
    LUNCH : _('Lunch'),
    TELEPHONE : _('On the phone'),
    BE_RIGHT_BACK : _('Be right back')}

REVERSE = dict([(y, x) for x, y in STATUS.items()])

def is_valid(status):
    '''return True if status is a valid status value'''

    return status in STATUS
        



# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import gtk
import gtk.gdk
import Theme
import InkDraw
import time

# Ink draw demo written by alencool

class InkDrawDemo(gtk.Window):
    def __init__(self, theme):
        gtk.Window.__init__(self)
        self.set_size_request(-1, 200)
        ### Drawing Area
        self._inkcanvas = InkDraw.InkCanvas( \
            pointer_pixbuf= theme.getImage('cursor-mouse'), 
            paintbrush_pixbuf= theme.getImage('tool-paintbrush'), 
            erasor_pixbuf= theme.getImage('tool-eraser') )

        ### paint button
        paint_image = gtk.Image()
        paint_image.set_from_pixbuf(theme.getImage('paintbrush'))
        paint_button = gtk.RadioToolButton()
        paint_button.set_icon_widget(paint_image)
        paint_button.connect('clicked', self.__set_paint)
        
        ### eraser button
        erase_image = gtk.Image()
        erase_image.set_from_pixbuf(theme.getImage('eraser'))
        erase_button = gtk.RadioToolButton(paint_button)
        erase_button.set_icon_widget(erase_image)
        erase_button.connect('clicked', self.__set_eraser)

        ### Color select
        self._color_label = InkDraw.ColorLabel(self._inkcanvas.get_stroke_color())
        hbox = gtk.HBox()        
        hbox.pack_start(self._color_label)
        hbox.pack_start(gtk.Arrow(gtk.ARROW_DOWN, gtk.SHADOW_IN))
        color_button = InkDraw.ColorButton(hbox)
        color_button.connect('color-set', self.__color_set, paint_button)
        
        ### clear button
        imgclear = gtk.Image()
        imgclear.set_from_stock(gtk.STOCK_CLEAR, gtk.ICON_SIZE_LARGE_TOOLBAR )
        clear = gtk.ToolButton(imgclear)
        clear.connect( 'clicked', self.__clear_canvas )
        
        
        main_img = gtk.Image()
        main_img.set_from_pixbuf(theme.getImage('medium_grid'))
        blank_img = gtk.Image()
        blank_img.set_from_pixbuf(theme.getImage('no_grid') )
        small_img = gtk.Image()
        small_img.set_from_pixbuf(theme.getImage('small_grid'))
        medium_img = gtk.Image()
        medium_img.set_from_pixbuf(theme.getImage('medium_grid'))
        large_img = gtk.Image()
        large_img.set_from_pixbuf(theme.getImage('large_grid'))
        
        ### grid button
        grid_button = InkDraw.GridButton(
            main_img, blank_img, small_img, medium_img, large_img,
            self._inkcanvas.get_grid_type())
        grid_button.connect( 'grid-type-set', self.__set_grid_type)

        ### undo Button
        undo_button = gtk.ToolButton()
        undo_button.set_stock_id(gtk.STOCK_UNDO)
        undo_button.connect("clicked", self.__undo)

        ### redo Button
        redo_button = gtk.ToolButton()
        redo_button.set_stock_id(gtk.STOCK_REDO)
        redo_button.connect("clicked", self.__redo)
        
        ### size Slider
        brushsmall = gtk.Image()
        brushsmall.set_from_pixbuf(theme.getImage('brush-small'))
        brushlarge = gtk.Image()
        brushlarge.set_from_pixbuf(theme.getImage('brush-large'))
        brush_slider = gtk.HScale(gtk.Adjustment(value=8, lower=1, upper=20, step_incr=0.5, page_incr=1, page_size=0))
        brush_slider.set_draw_value(True)
        brush_slider.set_size_request(130, -1)
        brush_slider.set_value_pos(gtk.POS_RIGHT)
        brush_slider.set_digits(0)
        brush_slider.set_update_policy(gtk.UPDATE_DISCONTINUOUS)
        brush_slider.connect("value-changed", self.__set_size)
                
        ### hbox
        hbox = gtk.HBox()
        hbox.pack_start(color_button, False, False)
        hbox.pack_start(paint_button, False, False)
        hbox.pack_start(erase_button, False, False)
        hbox.pack_start(undo_button, False, False)
        hbox.pack_start(redo_button, False, False) 
        hbox.pack_start(clear, False, False)
        hbox.pack_start(grid_button, False, False)
        hbox.pack_start(brushsmall, False, False)
        hbox.pack_start(brush_slider, False, False)
        hbox.pack_start(brushlarge, False, False) 
        
        ### vbox
        vbox = gtk.VBox()
        vbox.pack_start(hbox, expand= False)
        vbox.pack_start(self._inkcanvas)
        self.add(vbox)
        self.show_all()

    def __color_set(self, widget, color, paint_button):
        self._color_label.set_color(color)
        self._inkcanvas.set_stroke_color(color)
        self._inkcanvas.set_tool_type(InkDraw.ToolType.Paintbrush)
        paint_button.set_active(True)

    def __set_paint(self, widget=None):
        self._inkcanvas.set_tool_type(InkDraw.ToolType.Paintbrush)

    def __set_eraser(self, widget=None):
        self._inkcanvas.set_tool_type(InkDraw.ToolType.Eraser)

    def __clear_canvas(self, widget):
        self._inkcanvas.clear_canvas()

    def __set_grid_type(self, widget, grid_type):
        self._inkcanvas.set_grid_type(grid_type)

    def __undo(self, widget):
        self._inkcanvas.undo()
        
    def __redo(self, widget):
        self._inkcanvas.redo()
 
    def __set_size(self, range_widget):
        self._inkcanvas.set_stroke_size(range_widget.get_value())

if __name__ == "__main__":
    theme = Theme.Theme(None)
    win = InkDrawDemo(theme)
    win.connect('destroy', gtk.main_quit)
    gtk.main()



# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import gtk
import pango

import dialog

from emesenecommon import sort

class PluginManagerDialog( gtk.Window ):
    '''This class show and manage all the plugins installed
    in the plugins directory'''
    
    def __init__( self, pluginManager, parent, config ):
        '''Constructor'''
        
        gtk.Window.__init__( self )

        self.config = config
        
        self.pluginManager = pluginManager
        self.set_transient_for( parent )
        self.set_type_hint( gtk.gdk.WINDOW_TYPE_HINT_DIALOG )
        
        self.set_default_size( 500, 350 )
        self.set_geometry_hints( self, 500, 350, 500, 350 )
        self.set_title( _( 'Plugin Manager' ) )
        self.set_role( _( 'Plugin Manager' ) )

        self.set_border_width( 12 )
        self.set_resizable( False )
        
        vbox = gtk.VBox()
        vbox.set_spacing( 12 )
        
        self.listStore = gtk.ListStore( str, str, str, bool )
        
        cellRendererTextName = gtk.CellRendererText()
        cellRendererToggle = gtk.CellRendererToggle()
        cellRendererTextDesc = gtk.CellRendererText()
        
        cellRendererToggle.set_property( 'activatable', True )
        
        plugOn = gtk.TreeViewColumn( _( 'Active' ), cellRendererToggle, active=3 )
        plugOn.set_resizable( False )
        plugName = gtk.TreeViewColumn( _( 'Name' ), cellRendererTextName, markup=0 )
        plugName.add_attribute( cellRendererTextName, 'text', 0 )
        plugName.set_resizable( True )
        plugName.set_min_width( 60 )
        plugDesc = gtk.TreeViewColumn( _( 'Description' ), cellRendererTextDesc )
        cellRendererTextDesc.set_property( 'ellipsize', pango.ELLIPSIZE_END )
        plugDesc.add_attribute( cellRendererTextDesc, 'text', 2 )
        plugDesc.set_expand( True )
        plugDesc.set_sizing( gtk.TREE_VIEW_COLUMN_FIXED )
        
        self.list = gtk.TreeView( self.listStore )
        self.list.append_column( plugOn )
        self.list.append_column( plugName )
        self.list.append_column( plugDesc )
        self.list.set_reorderable( False )
        self.list.set_headers_visible( True )
        self.list.set_rules_hint( True )
        
        self.scroll = gtk.ScrolledWindow()
        self.scroll.set_policy( gtk.POLICY_NEVER , gtk.POLICY_AUTOMATIC )
        self.scroll.set_shadow_type( gtk.SHADOW_OUT )
        self.scroll.add( self.list )
        
        table = gtk.Table( 3, 2 )
        table.set_row_spacings( 6 )
        table.set_col_spacings( 12 )
        hButBox = gtk.HButtonBox()
        hButBox.set_spacing( 6 )
        hButBox.set_layout( gtk.BUTTONBOX_END )
                
        self.description = gtk.Label() 
        self.author = gtk.Label() 
        self.web = gtk.Label() 
        
        self.description.set_alignment( 0.0, 0.0 )
        self.author.set_alignment( 0.0, 0.5 )
        self.web.set_alignment( 0.0, 0.5 )
        
        self.description.set_line_wrap( True )
        self.author.set_line_wrap( True )
        self.web.set_line_wrap( True )
        
        self.description.set_size_request( 476, 48 )
        
        auth = gtk.Label( '<b>' + _( 'Author:' ) + '</b>' )
        auth.set_alignment( 0.0, 0.5 )
        auth.set_use_markup( True )
        web = gtk.Label( '<b>' + _( 'Website:' ) + '</b>' )
        web.set_alignment( 0.0, 0.5 )
        web.set_use_markup( True )
        
        table.attach( self.description , 0, 2, 0, 1, gtk.FILL | gtk.EXPAND )
        
        table.attach( auth, 0, 1, 1, 2, gtk.FILL )
        table.attach( self.author , 1, 2, 1, 2 )
        
        table.attach( web, 0, 1, 2, 3, gtk.FILL )
        table.attach( self.web , 1, 2, 2, 3 )
        
        self.bConfigure = gtk.Button( _( 'Configure' ), gtk.STOCK_PROPERTIES )
        self.bReload = gtk.Button( stock=gtk.STOCK_REFRESH )
        self.bclose = gtk.Button( stock=gtk.STOCK_CLOSE )

        self.bConfigure.connect( 'clicked', self.configurePlugin )
        self.bReload.connect( 'clicked', self.reloadPlugin )
        self.bclose.connect( 'clicked', self.close )

        hButBox.pack_start( self.bConfigure )
        hButBox.pack_start( self.bReload )
        hButBox.pack_start( self.bclose )
                        
        vbox.pack_start( self.scroll, True, True )
        vbox.pack_start( table, False, False )
        vbox.pack_start( hButBox, False, False )
       
        vbox.show_all()
        
        self.add( vbox )
        
        self.fillList()
        self.list.columns_autosize()
        
        self.connect( 'delete-event', self.close )
        self.list.connect( 'cursor-changed', self.row_selected )
        cellRendererToggle.connect( 'toggled', self.active_toggled, \
            ( self.listStore, 3 ) )
    
    def fillList( self ):
        '''fills the plugin list'''
        self.listStore.clear()
        
        for i in sort( self.pluginManager.getPlugins() ):
            plugin = self.pluginManager.getPlugin( i )
            if plugin == None:
                continue
            self.listStore.append( [ '<b>' + plugin.displayName + '</b>', i, \
                plugin.description, plugin.isEnabled() ] )
        
    def close( self, *args ):
        '''close the window'''
        
        self.destroy()
        
    def getSelected( self ):
        '''Return the selected item'''
        model = self.list.get_model()
        selected = self.list.get_selection().get_selected()
        if selected[1]:
            return model.get(selected[1], 1)[0]
        else:
            return None
        
    def row_selected( self, *args ):
        '''callback for the row_selected event'''
        
        plugin = self.pluginManager.getPlugin( self.getSelected() )
        self.setDescription( plugin )
        if hasattr(plugin, "configure") and callable(plugin.configure):
            self.bConfigure.set_sensitive(True)
        else:
            self.bConfigure.set_sensitive(False)
        
    def configurePlugin( self, *args ):
        '''call the configure method in the plugin'''
        plugin = self.pluginManager.getPlugin( self.getSelected() )
        if hasattr(plugin, "configure") and callable(plugin.configure):
            plugin.configure()
    
    def reloadPlugin( self, *args ):
        '''Reloads the plugin python code'''
        name = self.getSelected()
        plugin = self.pluginManager.getPlugin(name) 

        # save status and stop if needed
        wasEnabled = False
        if plugin != None and plugin.isEnabled():
            wasEnabled = True
            plugin.stop()

        # reload module
        module = self.pluginManager.pluginToModuleName(self.getSelected())
        self.pluginManager.loadPlugin(module, True)
        
        # start the plugin again
        plugin = self.pluginManager.getPlugin(name)
        if plugin != None and wasEnabled:
            plugin.start()
        
        # update the metadata
        self.fillList()
        self.setDescription(plugin)

    def setDescription( self, plugin ):
        if plugin != None:
            
            self.description.set_label( plugin.description )
            
            string = ''
            
            for (author,mail) in plugin.authors.iteritems():
                string += author + ' (' + mail + ')\n'
                
            
            self.author.set_label( '' + string[ :-1 ] )
            
            self.web.set_label( '' + plugin.website )
        
    def active_toggled( self, cell, path, user_data ):
        '''enable or disable the plugin'''
        
        model, column = user_data
        iterator= model.get_iter_from_string( path )
        
        plugin = self.pluginManager.getPlugin( model.get_value( iterator, 1 ) )
        
        if plugin != None:
            if plugin.isEnabled():
                plugin.stop()
                model.set_value( iterator, column, False )
        
        plugins = self.config.user['activePlugins'].split( ',' )
        
        if plugin.name in plugins:
            plugins.pop( plugins.index( plugin.name ) )
            self.config.user['activePlugins'] = ','.join(plugins)
        else:
            status = plugin.check()
                
            if status[ 0 ]:
                plugin.start()
                model.set_value( iterator, column, True )
            
                plugins = self.config.user['activePlugins'].split( ',' )
                
                if plugins[ 0 ] == '' and len( plugins ) == 1:
                    self.config.user['activePlugins'] = plugin.name
                elif not plugin.name in plugins:
                    s = ','.join(plugins) + ',' + plugin.name
                    self.config.user['activePlugins'] = s
            else:
                dialog.error(_('Plugin initialization failed: \n' ) \
                    + status[1])
                    



# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import gettext

class CurrentSong(object):

    customConfig = {}
    def __init__( self ):
        self.playing = ''
        self.artist = ''
        self.title = ''
        self.album = ''
        self.filename = ''
        
        style = '%title - %artist (%album)'
        
        self.setStyle( style )
        
        self.dictCommand = {
            'show': (self.cmd_Show,'Show playing song',True)
        }

        self._log = []
        self.status = 'unknown'

    def log(self, type, message):
        print "%s: %s" % (type, message)
        self._log.append((type, message))
    
    def getSongDict(self):
        songinfo = {}
        songinfo['artist'] = self.artist 
        songinfo['title'] = self.title
        songinfo['album'] = self.album
        return songinfo

    def getCurrentSong( self ):
        '''return the formated current song'''
        return self.parseStyle()
        
    def parseStyle( self ):
        '''return a parsed style according to the value of the variables'''
        if self.title == '' and self.artist == '' and self.album == '':
            return ''
        else:
            return self.style.replace('%artist', self.artist)\
                .replace('%title', self.title).replace('%album', self.album)
        
    def check( self ):
        '''check if there was a change in the song'''
        return False
        
    def isPlaying( self ):
        '''check if the player is playing'''
        return False

    def isRunning( self ):
        '''check if the player is running'''
        return False
        
    def getStatus( self ):
        '''check if everything is OK to start the plugin
        return a tuple whith a boolean and a message
        if OK -> ( True , 'some message' )
        else -> ( False , 'error message' )'''
        
        return ( True, 'OK' )
    
    def setStyle( self, string ):
        '''set the style'''
        self.style = '\\0Music\\01\\0' + string.replace('%title', '{0}').replace('%artist', '{1}').replace('%album', '{2}') + '\\0%title\\0%artist\\0%album\\0\\0'
        
    def is_on_path(self, fname):
        for p in os.environ['PATH'].split(os.pathsep):
            if os.path.isfile(os.path.join(p, fname)):
                return True
    
    # for plugins that need connecting/disconnecting signals
    def start( self ):
        pass
        
    def stop( self ):
        pass

    def getCoverPath( self ):
        return None
    
    def updateConfig( self ):
        pass
    
    def cmd_Show( self , *args):
        if self.artist == '' and self.album == '' and self.title == '':
            return ( False , 'Not Playing' )
        cm =  self.getCurrentSong()
        cm = cm[cm.find( '\\0Music\\01\\0')+12:]
        cmargs = cm.split('\\0')
        cm = cmargs[0]
        for args in range(1, len(cmargs)):
            cm = cm.replace( '{%s}' %str(args-1), cmargs[args])
        if cm == '':
            return ( False , 'Not Playing' )

        return ( True, cm )
    
ROOT_NAME = 'org.freedesktop.DBus'
ROOT_PATH = '/org/freedesktop/DBus'

DBUS = False

class DbusBase( CurrentSong ):
    
    def __init__( self, name = '', callback = None ):
        CurrentSong.__init__( self )
         
        global DBUS
        
        # http://listen-project.org/browser/trunk-0.6/src/dbus_manager.py#L29
        try:
            import dbus
            dbus_version = getattr(dbus, 'version', (0,0,0))
            if dbus_version >= (0,41,0) and dbus_version < (0,80,0):
                dbus.SessionBus()
                import dbus.glib
            elif dbus_version >= (0,80,0):
                from dbus.mainloop.glib import DBusGMainLoop
                DBusGMainLoop(set_as_default=True)
                dbus.SessionBus()
            else:
                self.log('error', 'python-dbus is too old!')
                raise
        except Exception, e:
            self.log('error', 'cant start dbus')
            DBUS = False
        else:
            DBUS = True
        
        if not DBUS:
            return
        self.module = dbus
        self.bus = dbus.SessionBus()
        self.root = self.bus.get_object( ROOT_NAME, ROOT_PATH )
        
        self.isNocWaiting = False
        if name and callback:
            self.reset( name, callback )
        
    def reset( self, name, callback ):
        self.log( 'info', 'reset player: ' + str(name) )
        self.status = 'not running'
        if self.isNameActive( name ):
            self.log( 'info', 'player running: ' + str(name) )
            self.status = 'running'
            callback()
        else:
            self.log( 'info', 'not running, listening NameOwnerChanged' )
            self.status = 'not running'
            def noc(changedName, *args):
                if str(changedName) == name and self.isNameActive(name):
                    self.log( 'info', 'player running: ' + str(name) )
                    self.status = 'running'
                    callback()
                    self.isNocWaiting = False
            # noc == name owner changed
            self.isNocWaiting = True
            self.bus.add_signal_receiver( noc, 'NameOwnerChanged', \
                                          ROOT_NAME, None, ROOT_PATH )
        
    def setCurrentSongData( self ):
        self.artist = ''
        self.title = ''
        self.album = ''
        self.filename = ''

    def getStatus( self ):
        '''don't override this'''
        
        global DBUS
        
        if os.name != 'posix':
            return ( False, _( 'This plugin only works in posix systems' ) )
        
        if not DBUS:
            return ( False, _( 'D-Bus cannot be initialized' ) )
        
        return ( True, 'Ok' )

    def isPlaying( self ):
        return False
        
    def check( self ):
        return False
    
    def isNameActive( self, name ):
        '''a helper for your class, so don't override it'''
        return bool( self.root.NameHasOwner(name) )
    




#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import re
import sys

from htmlentitydefs import codepoint2name

#
# Global variables for use throughout the program
#
APP_NAME = 'emesene'
APP_VERSION = '1.0.1'
OS_NAME = os.name
DIR_SEP = os.sep

class PATH(object):
    '''a namespace for paths'''

    DIR_SEP = os.sep

    if hasattr(sys, "frozen"):
        APP_PATH = os.path.dirname(sys.executable)
    else:
        APP_PATH = os.path.abspath(os.path.dirname(__file__))
    
    HOME_DIR = os.path.expanduser('~')
    CONF_DIR_NAME = '.config' + DIR_SEP + 'emesene1.0'
    CONFIG_DIR = HOME_DIR + DIR_SEP + CONF_DIR_NAME
        
    THEME_HOME_PATH = CONFIG_DIR + DIR_SEP + 'themes'
    THEME_SYSTEM_WIDE_PATH = APP_PATH + DIR_SEP + 'themes'
    DEFAULT_THEME_PATH = THEME_SYSTEM_WIDE_PATH + DIR_SEP + 'default' + DIR_SEP

    PLUGINS_HOME = 'pluginsEmesene'
    PLUGINS_SYSTEM_WIDE = 'plugins_base'
    PLUGIN_SYSTEM_WIDE_PATH = APP_PATH + DIR_SEP + PLUGINS_SYSTEM_WIDE
    PLUGIN_HOME_PATH = CONFIG_DIR + DIR_SEP + PLUGINS_HOME
    
    SMILIES_SYSTEM_WIDE_PATH = APP_PATH + DIR_SEP + 'smilies'
    SMILIES_HOME_PATH = CONFIG_DIR + DIR_SEP + 'smilies'
    DEFAULT_SMILIES_PATH = SMILIES_SYSTEM_WIDE_PATH + DIR_SEP + 'default' + DIR_SEP
    
    CONVTHEMES_SYSTEM_WIDE_PATH = APP_PATH + DIR_SEP + 'conversation_themes'
    CONVTHEMES_HOME_PATH = CONFIG_DIR + DIR_SEP + 'conversation_themes'
    DEFAULT_CONVTHEMES_PATH = CONVTHEMES_SYSTEM_WIDE_PATH + DIR_SEP + 'default' + DIR_SEP
    
    LANG_PATH = APP_PATH + DIR_SEP + 'po'
    SOUNDS_PATH = APP_PATH + DIR_SEP + 'sound_themes'

# workaround for pluginsEmesene/__init__.py
PLUGINS_HOME = PATH.PLUGINS_HOME
PLUGIN_SYSTEM_WIDE_PATH = PATH.PLUGIN_SYSTEM_WIDE_PATH

# Other constants
RESPONSE_DELETE_ALIAS = 50
RESPONSE_NO_AVATAR = -50
MAX_MESSAGE_LENGTH = 1100

#Regex
urlRe = re.compile("(^|[^A-Za-z0-9])(https?\:\/\/|ftp\:\/\/|www\.)([A-Za-z0-9\$_\.\+!\*':\(\|\)@,;\/\?&%=~#^`\\\-]+)")
reUnencode = re.compile( '=\\?([^\\?]+)\\?Q\\?([^\\?]+)\\?=' )
reChar = re.compile( '=([A-z0-9]{2})' )

#
# Global functions
#

def htmlEncode(string):
    '''this method return a string whit html tags encoded'''
    for i,j in codepoint2name.iteritems():
        if i <= 256 and str(chr(i)) in string:
            string = string.replace(chr(i), j)
    return string

def rgbToHexa( color ):
    '''take a gtk.gdk.Color end returns a string with html way color. Eg.: #FFCC00'''
    
    red = color.red >> 8
    green = color.green >> 8
    blue = color.blue >> 8
    
    return '#'+'%02X%02X%02X' % ( red , green , blue )

def sort( list ):
    '''Sorts a list in alphabetical order without regard of the capitalization'''
    
    list = list[:]
    list.sort( key= lambda x: str.lower( x ) )
    return list
    
def getLinks( text ):
    links = ()
    text = re.sub( urlRe, _searchAnchors, text.replace('%','%%') )
    return text, links

def _searchAnchors( data ):
    s, w, url = data.groups()
    href = w+url
    if w == 'www.':
        href  = 'http://'+href
    return '%s'




# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

VERSION = '0.2'
IFACE_NAME = 'org.exaile.DBusInterface'
IFACE_PATH = '/DBusInterfaceObject'
    
import CurrentSong

class Exaile( CurrentSong.DbusBase ):
    '''Exaile Interface'''
    
    def __init__( self ):
        CurrentSong.DbusBase.__init__( self, IFACE_NAME, self.setInterface )
        
        try: self.iface
        except: self.iface = None
       
    def setInterface( self ):
        proxy_obj = self.bus.get_object( IFACE_NAME, IFACE_PATH )
        self.iface = self.module.Interface( proxy_obj, IFACE_NAME )
        
    def getCoverPath( self ):
        if self.iface:
            return self.iface.get_cover_path()
        else:
            return None
     
    def setCurrentSongData( self ):
        if self.iface:
            self.artist = self.iface.get_artist()
            self.title = self.iface.get_title()
            self.album = self.iface.get_album()

    def getVersion( self ):
        try:
            self.iface.get_version()
        except:
            return False
        
        return True
    
    def isPlaying( self ):
        if not self.getVersion():
            return False
            
        if self.iface.get_artist() != None and \
           self.iface.get_title() != None and \
           self.iface.get_album() != None:
            return True
        
        return False
        
    def check( self ):
        if not self.iface or not self.isNameActive(IFACE_NAME):
            return
        
        if self.artist != self.iface.get_artist() or \
           self.title != self.iface.get_title() or \
           self.album != self.iface.get_album():
            self.setCurrentSongData()
            return True
            
        return False




# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# signal code:
# self.bus.add_signal_receiver( self.on_playing_uri_changed, \
#     'playingUriChanged', PLAYER_NAME, None, PLAYER_PATH )
# self.bus.add_signal_receiver( self.on_playing_changed, \
#     'playingChanged', PLAYER_NAME, None, PLAYER_PATH)
# self.bus.remove_signal_receiver( ... )

# based on plugin by Alberto Talavera

VERSION = '0.2'

#HELL_NAME = 'OMGWTFBBQ'
OBJECT_NAME = 'org.gnome.Rhythmbox'
PLAYER_NAME = 'org.gnome.Rhythmbox.Player'
PLAYER_PATH = '/org/gnome/Rhythmbox/Player'
SHELL_NAME = 'org.gnome.Rhythmbox.Shell'
SHELL_PATH = '/org/gnome/Rhythmbox/Shell'

import CurrentSong

class Rhythmbox( CurrentSong.DbusBase ):
    '''Rhythmbox interface'''
    
    def __init__( self ):
        CurrentSong.DbusBase.__init__( self, OBJECT_NAME, self.setInterface )
        try: self.iface
        except: self.iface = None
        
        try: self.rbshell
        except: self.rbshell = None
        
        self.uri = ''
    
    def setInterface( self ):
        rbobj = self.bus.get_object( OBJECT_NAME, PLAYER_PATH )
        self.iface = self.module.Interface( rbobj, PLAYER_NAME )
        rbshellobj = self.bus.get_object( OBJECT_NAME, SHELL_PATH )
        self.rbshell = self.module.Interface( rbshellobj, SHELL_NAME )
        
    def setCurrentSongData( self, uri ):
        if not self.rbshell: return
        
        data = self.rbshell.getSongProperties( uri )
        self.title = data['title']
        self.artist = data['artist']
        self.album = data['album']
        self.uri = uri
    
    def check( self ):
        if not self.iface or not self.isNameActive(OBJECT_NAME):
            return
        
        uri = self.iface.getPlayingUri()
        if uri != self.uri:
            self.setCurrentSongData( uri )
            return True
        return False
    
    def isPlaying( self ):
        if not self.iface:
            return False
        else:
            return bool( self.iface.getPlaying() )



# -*- coding: utf-8 -*-
#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import gtk
import cairo
import gobject
from Widgets import image_surface_from_pixbuf, \
                    set_context_color, \
                    resize_image_surface
from math import pi

class ToolType(object):
    Paintbrush, Eraser = range(2)

class GridType(object):
    Blank, Small, Medium, Large = range(4)
    
class InkCanvas(gtk.DrawingArea):

    __gsignals__ = {
        'expose_event' : 'override',
        'motion_notify_event' : 'override',
        'button_press_event' : 'override',
        'button_release_event' : 'override',
        'realize': 'override',
        'configure-event': 'override',
        }

    _history = []
    _paths = []     
    _tool_type = ToolType.Paintbrush
    _stroke_size = 8
    _stroke_color = gtk.gdk.Color(0, 0, 0)
    _grid_type = GridType.Blank  
    _grid_size = {
        GridType.Blank:     -1,
        GridType.Small:     10,
        GridType.Medium:    20,
        GridType.Large:     30,
        }    
    _pointer_surface = None
    _cached_grid_surface = None
    _cached_painting_surface = None
    _tool_type_to_surface = {}
    _is_pressed = False
    _intermediate_path = None

    
    def __init__(self, pointer_pixbuf=None, paintbrush_pixbuf=None, erasor_pixbuf=None ):
        gtk.DrawingArea.__init__(self)
        self.set_events( 
            gtk.gdk.EXPOSURE_MASK       | gtk.gdk.LEAVE_NOTIFY_MASK     | gtk.gdk.BUTTON_PRESS_MASK |
            gtk.gdk.BUTTON_RELEASE_MASK | gtk.gdk.POINTER_MOTION_MASK   | gtk.gdk.POINTER_MOTION_HINT_MASK) 

        self._pointer_surface = image_surface_from_pixbuf(pointer_pixbuf)
        self._tool_type_to_surface = {
            ToolType.Paintbrush: image_surface_from_pixbuf(paintbrush_pixbuf),
            ToolType.Eraser: image_surface_from_pixbuf(erasor_pixbuf), 
            }
        
    def set_tool_type(self, tool_type=ToolType.Paintbrush):
        self._tool_type = tool_type
        self.__update_cursor()
        
    def set_stroke_size(self, stroke_size= 10):
        if stroke_size < 1:
            self._stroke_size = 1
        else:
            self._stroke_size = stroke_size
        self.__update_cursor()

    def set_stroke_color(self, color = gtk.gdk.Color(0, 0, 0)):
        self._stroke_color = color
    
    def get_stroke_color(self):
        return self._stroke_color
    
    def set_grid_type(self, grid_type):
        self._grid_type = grid_type
        self.__update_grid_surface(self.allocation.width, self.allocation.height)
        self.queue_draw()
    
    def get_grid_type(self):
        return self._grid_type
   
    def undo(self):
        if len(self._paths) > 0:
            self._history.append(self._paths.pop())
            self._cached_painting_surface = self.__build_paths_surface()
            self.queue_draw()
                    
    def redo(self):
        if len(self._history) > 0:
            self._paths.append(self._history.pop())
            self._cached_painting_surface = self.__build_paths_surface()
            self.queue_draw()

    def clear_canvas(self):
        self._cached_painting_surface = None
        self._paths=[]
        self._history = []
        self.queue_draw()
 
    def get_pixbuf(self):
        rec = gtk.gdk.Rectangle()
        for path in self._paths:
            rec = rec.union(path.get_path_rectangle())
        width = max(1, rec.width)
        height = max(1, self.allocation.height)
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        ctx = cairo.Context(surface)
        for path in self._paths:
            path.draw(ctx)
        return gtk.gdk.pixbuf_new_from_data(
            surface.get_data(), 
            gtk.gdk.COLORSPACE_RGB, True, 8, 
            surface.get_width(), surface.get_height(), 
            surface.get_stride())

    def do_button_press_event(self, event):
        if event.button == 1:
            self._is_pressed = True
            self._history = []
            path = BrushPath(Point(event.x, event.y), self._stroke_size, self._stroke_color, self._tool_type)
            self._paths.append(path)
            rec = path.get_intermediate_rectangle()
            self._intermediate_path = path
            self.queue_draw_area(rec.x, rec.y, rec.width, rec.height)
        return True
        
    def do_motion_notify_event(self, event):
        x,y =  event.get_coords()
        if self._is_pressed:
            path = self._paths[-1]
            path.add(Point(x,y))
            rec = path.get_intermediate_rectangle()
            self.queue_draw_area(rec.x, rec.y, rec.width, rec.height)
        return True

    def do_button_release_event(self, event):
        if self._is_pressed:
            self._is_pressed = False
            self._intermediate_path = None
            self.__update_cached_painting_surface(self._paths[-1])
        return True

    def do_expose_event(self, event):
        x , y, width, height = event.area
        ctx = event.window.cairo_create()      
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        cr = cairo.Context(surface)
        cr.translate(-x, -y)
        if self._cached_painting_surface:
            cr.set_source_surface(self._cached_painting_surface,0,0)        
            cr.paint()
        if self._intermediate_path:
            self._intermediate_path.draw(cr,0,0)
        self.__draw_grid(cr, x, y)
        self.__draw_canvas(cr, x,y, width, height)
        ctx.set_source_surface(surface,x,y)
        ctx.paint()
        return False

    def do_realize(self):
        gtk.DrawingArea.do_realize(self)
        self.__update_cursor()
        
    def do_configure_event(self, event):
        self.__update_grid_surface(event.width, event.height)
      
    def __draw_grid(self, ctx, x,y):
        if self._cached_grid_surface is not None:
            ctx.save()
            ctx.set_operator(cairo.OPERATOR_DEST_OVER)            
            ctx.set_source_surface(self._cached_grid_surface, 0, 0)
            ctx.paint()
            ctx.restore()

    def __draw_canvas(self, ctx, x,y, width, height, color = gtk.gdk.color_parse('#ffffff')):
        ctx.save()
        set_context_color(ctx, color)
        ctx.set_operator(cairo.OPERATOR_DEST_OVER)
        ctx.paint()
        ctx.restore()

    def __update_cached_painting_surface(self, path):

        if not self._cached_painting_surface:
            self._cached_painting_surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 1, 1)
        
        surface_width = self._cached_painting_surface.get_width()
        surface_height = self._cached_painting_surface.get_height()
        rec = path.get_path_rectangle() 
    
        combined = rec.union( gtk.gdk.Rectangle(0,0,surface_width,surface_height))        
        if (surface_width < combined.width) or (surface_height < combined.height):
            self._cached_painting_surface = resize_image_surface(
                self._cached_painting_surface, combined.width, combined.height)
        ctx = cairo.Context(self._cached_painting_surface)
        path.draw(ctx)
        self.queue_draw_area(rec.x, rec.y, rec.width, rec.height)
                
    def __update_grid_surface(self, width, height, color = gtk.gdk.color_parse('#C0C0C0')):
        self._cached_grid_surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        ctx = cairo.Context(self._cached_grid_surface)        
        space = self._grid_size[self._grid_type]
        if space < 1: return
        x, y = 0,0
        position = x + space
        end = x + width
        ctx.save()
        set_context_color(ctx, color)
        ctx.set_line_width(1)
        position = x + space
        end = x + width
        while position < end:
            ctx.move_to(position, y)
            ctx.line_to(position, y + height)
            position += space
        position = y + space
        end = y + height
        while position < end:
            ctx.move_to(x, position )
            ctx.line_to(x + width, position)
            position += space
        ctx.stroke()
        ctx.restore()

    def __build_paths_surface(self):
        rec = gtk.gdk.Rectangle()
        for path in self._paths:
            rec = rec.union(path.get_path_rectangle())
        width = max(1, rec.width)
        height = max(1, rec.height)
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        ctx = cairo.Context(surface)
        for path in self._paths:
            path.draw(ctx)
        return surface

    def __update_cursor(self):
        indicator_surface = self._tool_type_to_surface[self._tool_type]
        radius = int(self._stroke_size/2.0)
        shift = radius +1
        cursor_width = int(radius + max(self._pointer_surface.get_width(),  indicator_surface.get_width()))
        cursor_height = int(radius + max(self._pointer_surface.get_height(), indicator_surface.get_height()))                    
        cursor_surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, cursor_width, cursor_height)
        ctx = cairo.Context(cursor_surface)
        ctx.save()
        ctx.set_line_width(1.0), ctx.set_source_rgb(0,0,0), ctx.set_dash([1,1],0)
        ctx.arc(shift,shift,radius, 0, 2 * pi), ctx.stroke()
        ctx.set_source_surface(self._pointer_surface, shift, shift), ctx.paint()
        ctx.set_source_surface(indicator_surface, shift, shift), ctx.paint()
        ctx.restore()
        pixbuf = gtk.gdk.pixbuf_new_from_data(
            cursor_surface.get_data(), 
            gtk.gdk.COLORSPACE_RGB, True, 8, 
            cursor_surface.get_width(), cursor_surface.get_height(), 
            cursor_surface.get_stride())
        cursor = gtk.gdk.Cursor(self.window.get_display(), pixbuf, shift, shift)
        self.window.set_cursor(cursor)


class BrushPath(object):
    """ represents a single brush path and has the ability to draw itself """
    def __init__(self, start_point, stroke_size, color=gtk.gdk.Color(0,0,0), tool_type = ToolType.Paintbrush):
        self._points = []
        self._stroke_size = stroke_size
        self._shift = int((stroke_size/2) + 1)
        self._color = color
        self._tool_type = tool_type
        if tool_type == ToolType.Paintbrush:
            self._operator = cairo.OPERATOR_OVER
        else: # tool_type == ToolType.Eraser
            self._operator = cairo.OPERATOR_CLEAR
        self._path_rectangle = gtk.gdk.Rectangle()
        self._intermediate_rectangle = gtk.gdk.Rectangle() 
        self.add(start_point)

    def add(self, point):
        self._points.append(point)                    
        #--- update path rectangle
        i_rec = self.__create_point_rectangle(point)
        self._path_rectangle = self._path_rectangle.union(i_rec)
        
        #... update intermediate rectangle            
        i_points = self._points[-2:len(self._points)]
        for pnt in i_points[0:len(i_points)-1]:
            z_rec = self.__create_point_rectangle(pnt)                
            i_rec = i_rec.union(z_rec)
        self._intermediate_rectangle = i_rec

    def get_path_rectangle(self):
            return self._path_rectangle
    def get_intermediate_rectangle(self):
            return self._intermediate_rectangle

    def draw(self, ctx, x=0, y=0):
        points = self._points
        if not points: return
        ctx.save()
        ctx.set_operator(self._operator)
        ctx.translate(x,y)
        x, y, width, height = self._path_rectangle
        ctx.rectangle(x,y, width, height)
        ctx.clip()
        ctx.set_antialias(cairo.ANTIALIAS_SUBPIXEL)   
        ctx.set_line_cap(cairo.LINE_CAP_ROUND)
        ctx.set_line_join(cairo.LINE_JOIN_ROUND)
        set_context_color(ctx, self._color)
        ctx.set_line_width(self._stroke_size)
        ctx.move_to(points[0].x, points[0].y)
        for point in points[1:len(points)]:
            ctx.line_to(point.x, point.y)
        if len(points) == 1:
            ctx.line_to(points[0].x, points[0].y)
        ctx.stroke()
        ctx.restore()
                
    def __create_point_rectangle(self, point):
        return gtk.gdk.Rectangle( 
            int(point.x - self._shift), int(point.y - self._shift), 
            int(self._shift * 2), int(self._shift * 2))   


class ColorLabel(gtk.Widget):
    ''' Square shaped, color label '''   
    __gsignals__ = { 'size_request' : 'override', 'expose-event' : 'override' }
    
    _color = gtk.gdk.color_parse('#000000')
    _dimention = 20
    
    def __init__(self, color = gtk.gdk.color_parse('#000000')):
        gobject.GObject.__init__(self)
        gtk.Widget.__init__(self)
        self.set_flags(self.flags() | gtk.NO_WINDOW )       
        self._color = color

    def do_size_request(self,requisition):
        requisition.width = self._dimention
        requisition.height = self._dimention

    def do_expose_event(self, event):
        x , y, width, height = event.area
        dimention = min(width,height)
        y += (height /2) - (dimention /2)
        width, height = dimention, dimention
        ctx = event.window.cairo_create()
        ctx.translate(x, y) 
        ctx.set_antialias(cairo.ANTIALIAS_SUBPIXEL)        
        ctx.rectangle(0, 0, width, height)
        ctx.clip_preserve()
        set_context_color(ctx, self._color)
        ctx.fill_preserve()
        ctx.set_line_width(1.0)
        set_context_color(ctx, gtk.gdk.color_parse('#000000'))
        ctx.stroke()

    def set_color(self, color=gtk.gdk.color_parse('#000000')):
        self._color = color
        self.queue_draw()
        
    def get_color(self): 
        return self._color

gobject.type_register(ColorLabel)

class ColumnContainer(gtk.Box):
    ''' Container that postions widgets into a given column number'''
    __gsignals__ = { 'size_request' : 'override','size_allocate' : 'override' }
    
    _cols = 1       # number of columns
    _cr_width = 0   # child requisition width
    _cr_height = 0  # child requisition height
    _vc_number = 0  # number of visual children
    
    def __init__(self, column_number=1):
        gtk.Box.__init__(self)
        if column_number < 1:
            self._cols = 1
        else:
            self._cols = column_number

    def update_child_info(self):
        visual_children_number = 0
        child_requisition_width = 0
        child_requisition_height = 0
        for child in self.get_children():
            if child.get_property('visible'):
                child_request_width, child_request_height = child.size_request()
                child_requisition_width = max(child_requisition_width, child_request_width)
                child_requisition_height = max(child_requisition_height, child_request_height)
                visual_children_number += 1
        self._cr_width = child_requisition_width
        self._cr_height = child_requisition_height
        self._vc_number = visual_children_number
                        
    def do_size_request(self, requisition):
        self.update_child_info()        
        width, height = 0, 0          
        if self._vc_number > 0:
            width = self._cr_width * self._cols
            if self._vc_number <= self._cols:    
                height = self._cr_height
            elif (self._vc_number % self._cols ) == 0 :
                height = self._cr_height * ( self._vc_number / self._cols)
            else:
                height = self._cr_height * (( self._vc_number / self._cols) + 1)
        requisition.width = width
        requisition.height = height

    def do_size_allocate(self, allocation):     
        width = allocation.width / self._cols
        height = self._cr_height
        x, y = allocation.x, allocation.y
        col_spaces_remaining = self._cols
        for child in self.get_children():
            child.size_allocate(gtk.gdk.Rectangle(x, y, width, height))
            col_spaces_remaining -= 1
            if col_spaces_remaining == 0:
                y += height
                x = allocation.x
                col_spaces_remaining = self._cols
            else:
                x += width
                
gobject.type_register(ColumnContainer)

class DropdownWindow(gtk.Window):
    __gsignals__ = {
        'map-event' : 'override',
        'unmap-event' : 'override',
        'button-press-event' : 'override',
        'key-press-event' : 'override',
        }

    def __init__(self):
        gtk.Window.__init__(self, gtk.WINDOW_POPUP)
        self.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_MENU)
        self.set_decorated(False)

    def do_map_event(self, event):
        self.grab_focus()
        self.grab_add()
        time = gtk.get_current_event_time()
        gtk.gdk.pointer_grab(self.window, True, gtk.gdk.BUTTON_PRESS_MASK, None, None, time)
        gtk.gdk.keyboard_grab(self.window, True, time)

    def do_unmap_event(self, event): 
        time = gtk.get_current_event_time()
        gtk.gdk.pointer_ungrab(time)
        gtk.gdk.keyboard_ungrab(time)
        self.grab_remove()

    def do_button_press_event(self, event):
        win_tuple = gtk.gdk.window_at_pointer()
        if (win_tuple is None) or not( win_tuple[0] == self.window) : self.hide()
        return True

    def do_key_press_event(self, event):
        if not gtk.Window.do_key_press_event(self, event) \
               and event.keyval == gtk.gdk.keyval_from_name("Escape"):
            self.hide()
            return True

gobject.type_register(DropdownWindow)

class ColorButton(gtk.ToggleButton):
    __gsignals__ = { 
        'color-set' : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,(gobject.TYPE_PYOBJECT,)),
        'button-press-event' : 'override'
        }
    _color = gtk.gdk.Color(0,0,0)
    _tango_colors = [ \
        '#2e3436', '#babdb6', '#a40000', '#5c3566', '#204a87', '#4e9a06', '#8f5902', '#ce5c00', '#c4a000',
        '#555753', '#d3d7cf', '#cc0000', '#75507b', '#3465a4', '#73d216', '#c17d11', '#f57900', '#edd400',
        '#888a85', '#eeeeec', '#ef2929', '#ad7fa8', '#729fcf', '#8ae234', '#e9b96e', '#fcaf3e', '#fce94f']

    def __init__(self, widget):
        gtk.Button.__init__(self)
        self.add(widget)
        self.set_property("can-default", False)
        self.set_property("can-focus", False)
        self.set_border_width(0)
        self.set_relief(gtk.RELIEF_NONE)
        self._dropdown_win = DropdownWindow()
        self._dropdown_win.realize()
        self._dropdown_win.connect("map-event", self.__win_map)
        self._dropdown_win.connect("unmap-event", self.__win_unmap)

        container = ColumnContainer(9)
        hbox = gtk.HBox()
        for hex_color in self._tango_colors:       
            container.add(self.__create_button(hex_color))
        for hex_color in ('#000000','#ffffff'):
            hbox.pack_start(self.__create_button(hex_color), False, False)    
        vbox = gtk.VBox()
        vbox.pack_start(hbox, False, False)
        vbox.pack_start(container, False, False)
        self._dropdown_win.add(vbox)
        
    def __create_button(self, hex_color='#000000'):
        label = ColorLabel(gtk.gdk.color_parse(hex_color))
        button = gtk.Button()
        button.set_relief(gtk.RELIEF_NONE)
        button.set_property("can-default", False)
        button.set_property("can-focus", False)                
        button.set_border_width(0)
        button.add(label)
        button.connect('clicked', self.__color_selected, label.get_color())    
        return button
        
    def __color_selected(self, widget, color):
        self._color = color
        self._dropdown_win.hide()
        self.emit('color-set', self._color)
    
    def __win_map(self, win, event): self.set_active(True)
    def __win_unmap(self, win, event): self.set_active(False)

    def __position(self, win):
        x, y = self.window.get_origin()
        alloc = self.allocation
        x, y = x + alloc.x, y + alloc.y
        width, height = win.get_size()
        if x + width > self.get_screen().get_width(): x += alloc.width - width
        if y + alloc.height + height > self.get_screen().get_height(): y -= height
        else: y += alloc.height
        win.move(x, y)
    
    def get_color(self): return self._color
    
    def do_button_press_event(self, event):
        if not self.get_active():
            self.__position(self._dropdown_win)
            self._dropdown_win.show_all()
            self._dropdown_win.present()
        return True        
gobject.type_register(ColorButton)

class GridButton(gtk.ToggleButton):
    __gsignals__ = { 
        'grid-type-set' : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,(gobject.TYPE_PYOBJECT,)),
        'button-press-event' : 'override'
        }

    def __init__(self, button_icon, blank_icon, small_icon, medium_icon, large_icon, grid_type=GridType.Blank):
        gtk.Button.__init__(self)
        hbox = gtk.HBox()
        hbox.pack_start(button_icon)
        hbox.pack_start(gtk.Arrow(gtk.ARROW_DOWN, gtk.SHADOW_IN))
        self.add(hbox)
        self.set_property("can-default", False)
        self.set_property("can-focus", False)
        self.set_border_width(0)
        self.set_relief(gtk.RELIEF_NONE)
        self._grid_type = grid_type
        
        ### Grid Menu
        self._blank_item = self.__create_grid_item('Blank', blank_icon, GridType.Blank)
        self._small_item = self.__create_grid_item('Small grid', small_icon, GridType.Small)
        self._medium_item = self.__create_grid_item('Medium grid', medium_icon, GridType.Medium)
        self._large_item = self.__create_grid_item('Large grid', large_icon, GridType.Large)
        
        self._grid_menu = gtk.Menu()
        self._grid_menu.append(self._blank_item)
        self._grid_menu.append(gtk.SeparatorMenuItem())
        self._grid_menu.append(self._small_item)        
        self._grid_menu.append(self._medium_item)
        self._grid_menu.append(self._large_item)
        self._grid_menu.connect("unmap-event", self.__win_unmap)           
    
    def __create_grid_item(self, label, icon, grid_type):
        grid_item = gtk.ImageMenuItem(label)
        grid_item.set_image(icon)
        grid_item.connect('activate', self.__item_activate, grid_type)
        return grid_item
        
    def __item_activate(self, menuitem, grid_type):
        self._grid_type = grid_type
        self.emit('grid-type-set', self._grid_type)
        
    def __win_unmap(self, win, ev): self.set_active(False)
        
    def __menu_postion(self, menu):
        x, y = self.window.get_origin()
        alloc = self.allocation                
        x, y = x + alloc.x, y + alloc.y
        menu_allocation = menu.allocation
        width, height = menu_allocation.width, menu_allocation.height
        if x + width > self.get_screen().get_width(): x += alloc.width - width
        if y + alloc.height + height > self.get_screen().get_height(): y -= height
        else: y += alloc.height
        return (x,y, True)
        
    def do_button_press_event(self, event):
        if not self.get_active():
            self._grid_menu.popup(None, None, self.__menu_postion, event.button, event.time)
            self.set_active(True)
            self._grid_menu.show_all()
            if self._grid_type == GridType.Blank: self._grid_menu.select_item(self._blank_item)
            elif self._grid_type == GridType.Small: self._grid_menu.select_item(self._small_item)
            elif self._grid_type == GridType.Medium: self._grid_menu.select_item(self._medium_item)
            elif self._grid_type == GridType.Large: self._grid_menu.select_item(self._large_item)                    
        return True     
           
gobject.type_register(GridButton)

class Point(object):
    def __init__(self,x,y):
        self.x = x
        self.y = y
    
    def __eq__(self, other):
        if (self.x == other.x) and (self.y == other.y):
            return True
        else:
            return False
